" Vimball Archiver by Charles E. Campbell, Jr., Ph.D.
UseVimball
finish
plugin/VimTagsManager.vim	[[[1
100
" Vim script utilities for VimLite
" Last Change: 2011 May 10
" Maintainer: fanhe <fanhed@163.com>
" License:  GPLv2

if exists('g:loaded_VimTagsManager')
    finish
endif
let g:loaded_VimTagsManager = 1

let s:start = 0

command! -nargs=* -complete=file VTMParseFiles :call g:VTMParseFiles(<f-args>)


function s:InitVariable(varName, defaultVal) "{{{2
    if !exists(a:varName)
		let {a:varName} = a:defaultVal
        return 1
    endif
    return 0
endfunction
"}}}

"call s:InitVariable('g:VimTagsManager_DbFile', 
            "\'~/Desktop/VimLite/CtagsDatabase/TestTags3.db')
call s:InitVariable('g:VimTagsManager_DbFile', 
            \'VimLiteTags.db')

call s:InitVariable('g:VimTagsManager_SrcDir', 
            \'~/.vim/autoload/omnicpp')

function! g:GetTagsByScopeAndKind(scope, kind) "{{{2
    py vim.command("let tags = %s" % vtm.GetTagsByScopeAndKind(
                \vim.eval('a:scope'), vim.eval('a:kind')))
    return tags
endfunction


function! g:GetTagsByScopesAndKinds(scopes, kinds) "{{{2
    py vim.command("let tags = %s" % vtm.GetTagsByScopesAndKinds(
                \vim.eval('a:scopes'), vim.eval('a:kinds')))
    return tags
endfunction


function! g:GetTagsByScopeAndName(scope, name) "{{{2
    py vim.command("let tags = %s" % vtm.GetTagsByScopeAndName(
                \vim.eval('a:scope'), vim.eval('a:name')))
    return tags
endfunction


function! g:GetTagsByScopesAndName(scopes, name) "{{{2
    py vim.command("let tags = %s" % vtm.GetTagsByScopeAndName(
                \vim.eval('a:scopes'), vim.eval('a:name')))
    return tags
endfunction

function! g:GetTagsByPath(path) "{{{2
    py vim.command("let tags = %s" % vtm.GetTagsByPath(vim.eval('a:path')))
    return tags
endfunction

function! g:GetTagsByKindAndPath(kind, path) "{{{2
    py vim.command("let tags = %s" % vtm.GetTagsByKindAndPath(
                \vim.eval('a:kind'), vim.eval('a:path')))
    return tags
endfunction

function! g:VTMParseFiles(...) "{{{2
    "若传进来的第一个参数为列表, 进解析此列表
    if a:0 > 0
        if type(a:1) == type([])
            py vtm.ParseFiles(vim.eval('a:1'))
            return
        endif
    endif

    py vtm.ParseFiles(vim.eval('a:000'))
endfunction


function! VimTagsManagerInit() "{{{1

python << PYTHON_EOF
# -*- encoding:utf-8 -*-
import sys, os, os.path
import vim

sys.path.extend([os.path.expanduser(vim.eval('g:VimTagsManager_SrcDir'))])
from VimTagsManager import VimTagsManager
PYTHON_EOF

    py vtm = VimTagsManager()
    py vtm.OpenDatabase(os.path.expanduser(vim.eval('g:VimTagsManager_DbFile')))
endfunction


" vim:fdm=marker:fen:expandtab:smarttab:fdl=1:
after/ftplugin/c.vim	[[[1
1
call omnicpp#complete#Init()
after/ftplugin/cpp.vim	[[[1
1
call omnicpp#complete#Init()
autoload/omnicpp/resolvers.vim	[[[1
1006
" Description:  Omnicpp completion resolving functions
" Maintainer:   fanhe <fanhed@163.com>
" Last Change:  2011 May 15
" License:      GPLv2

" 基本数据结构
"
" TypeInfo, 字典
" {
" 'name':   <name>, 
" 'tsl':    <template specialization list>, 
" }

" VarInfo, 字典
" {
" 'kind': <'container'|'variable'|'function'|'cast'|'unknown'>
" 'name': <name>
" 'typeInfo': <TypeInfo>
" 'tag': <tag>
" }

function! s:StripString(s) "{{{2
    let s = substitute(a:s, '^\s\+', '', 'g')
    let s = substitute(s, '\s\+$', '', 'g')
    return s
endfunc
"}}}
function! s:NewTypeInfo() "{{{2
    return {'name': '', 'tm': {}, 'tsl': [], 'tdl': []}
endfunc
"}}}
function! s:NewOmniScope() "{{{2
    return {'kind': 'unknown', 'name': ''}
endfunc
"}}}
" 获取全能补全请求前的语句的 OmniScopeStack
" 以下情况一律先清理所有不必要的括号
" case1. A::B C::D::|
" case2. A::B()->C().|    orig: A::B::C(" a(\")z ", '(').|
" case3. A::B().C->|
" case4. A->B().|
" case5. A->B.|
" case6. Z Y = ((A*)B)->C.|
" case7. (A*)B()->C.|
" case8. static_cast<A*>(B)->C.|
" case9. A(B.C()->|)
"
" 返回列表 omniSS, 每个条目为 omniScope
" {
" 'kind': <'container'|'variable'|'function'|'cast'|'unknown'>
" 'name': <name>
" }
"
" 如 case3, [{'kind': 'container', 'name': 'A'},
"            {'kind': 'function', 'name': 'B'},
"            {'kind': 'variable', 'name': 'C'}]
" 如 case6, [{'kind': 'cast', 'name': 'A*'},
"            {'kind': 'variable', 'name': 'B'},
"            {'kind': 'variable', 'name': 'C'}]
"
" TODO: 暂时无视 cast, 太复杂
" 判断 cast 的开始: 1. )单词, 2. )(
" 判断 precast: 从 )( 匹配的结束位置寻找匹配的 ), 如果匹配的 ')' 右边也为 ')'
" 判断 postcast: 从 )( 匹配的结束位置寻找匹配的 ), 如果匹配的 ')' 右边不为 ')'
function! omnicpp#resolvers#GetOmniScopeStack(...) "{{{2
    if a:0 > 0
        if type(a:1) == type('')
            let tokens = omnicpp#tokenizer#Tokenize(
                        \omnicpp#utils#SimplifyCodeForOmni(a:1))
        elseif type(a:1) == type([])
            let tokens = a:1
        else
            return []
        endif
    else
        let tokens = omnicpp#utils#TokenizeStatementBeforeCursor(1)
    endif

    let revTokens = reverse(tokens[:])

    let result = []

    " 反向遍历分析
    let idx = 0
    let revTokensLen = len(revTokens)

    " 0 -> 期望操作符 ('->', '.', '::')
    " 1 -> 期望单词 ( A|:: )
    " 2 -> 期望左括号 ( A(|). )
    let state = 0

    " 只有四种需要分辨的符号
    " 1. 作用域或成员操作符 ('::', '.', '->')
    " 2. 单词
    " 3. 左括号
    " 4. 右括号
    let prevToken = {}
    while idx < revTokensLen
        let token = revTokens[idx]

        if token.kind == 'cppOperatorPunctuator' && token.value =~ '->\|::\|\.'
            if state == 0
                " 期望操作符, 遇到操作符
                " 切换状态
                let state = 1 " 期望单词
            elseif state == 1
                " 期望单词, 遇到操作符
                " 语法错误
                echom 'syntax error: ' . token.value
                break
            elseif state == 2
                " 括号状态都已经私自处理完, 这里不再需要了
            endif
        elseif token.kind == 'cppWord' 
                    \|| (token.kind == 'cppKeyword' && token.value == 'this')
            if state == 0
                " 期望操作符, 遇到单词
                " 结束. eg A::B |C::
                break
            elseif state == 1
                " 成功获取一个单词
                let omniScope = s:NewOmniScope()
                let omniScope.name = token.value
                if prevToken.value == '::'
                    let omniScope.kind = 'container'
                elseif prevToken.value =~# '->\|\.'
                    let omniScope.kind = 'variable'
                else
                endif

                " 切换至期望操作符
                let state = 0

                let result = [omniScope] + result
            elseif state == 2
                " 括号状态都已经私自处理完, 这里不再需要了
            endif
        elseif token.kind == 'cppOperatorPunctuator' && token.value == '('
            " eg: A(B.C()->|)
            " 直接结束
            break
        elseif token.kind == 'cppOperatorPunctuator' && token.value == ')'
            if state == 0
                " 期待操作符
                " 遇到右括号
                " 必定是一个 postcast
                " 无须处理, 直接完成
                break
            elseif state == 1
                " 期待单词
                " 遇到右括号
                " 可能是 precast 或者 是一个函数
                " 判定是否函数
                if revTokens[idx+1].value == '(' 
                            \&& revTokens[idx+2].kind == 'cppWord'
                    " 是函数
                    let omniScope = s:NewOmniScope()
                    let omniScope.kind = 'function'
                    let omniScope.name = revTokens[idx+2].value
                    let result = [omniScope] + result

                    let idx = idx + 2
                    let token = revTokens[idx]
                    " 切换至期待操作符
                    let state = 0
                elseif revTokens[idx+1].kind == 'cppWord'
                    " 是一个 precast

                    " 添加变量 scope
                    let omniScope = s:NewOmniScope()
                    let omniScope.kind = 'variable'
                    let omniScope.name = revTokens[idx+1].value
                    let result = [omniScope] + result

                    " 直接处理完 precast. eg. ((A*)B)|
                    " 寻找匹配的左括号的位置
                    let j = idx + 2
                    let tmpToken = revTokens[j]
                    if tmpToken.value != ')'
                        echom 'syntax error: ' . token.value
                        break
                    endif
                    let j += 1
                    let tmp = 1
                    while j < revTokensLen
                        let tmpToken = revTokens[j]
                        if tmpToken.value == ')'
                            let tmp += 1
                        elseif tmpToken.value == '('
                            let tmp -= 1
                            if tmp == 0
                                break
                            endif
                        endif
                        let j += 1
                    endwhile

                    " TODO: 处理复杂的 cast 声明. eg. ((*)[10][20])A
                    let omniScope = s:NewOmniScope()
                    let omniScope.kind = 'cast'
                    let omniScope.name = revTokens[j-1].value
                    let result = [omniScope] + result

                    " 处理完毕
                    break
                else
                    echom 'syntax error: ' . token.value
                    break
                endif
            elseif state == 2
                " 括号状态都已经私自处理完, 这里不再需要了
            endif
        endif

        let prevToken = token
        let idx += 1
    endwhile

    return result
endfunc
"}}}
" 从 scope 和 name 组合生成 path, 处理烦人的 <global>
function! s:BuildPath(szScope, szName) "{{{2
    if a:szScope != '<global>'
        return a:szScope . '::' . a:szName
    else
        return a:szName
    endif
endfunc
"}}}
" 解析 omniSS, 返回用于获取 tags 的 omniTagScopes
function! omnicpp#resolvers#ResolveOmniScopeStack(scopeStack, omniSS) "{{{2
    if empty(a:scopeStack) || empty(a:omniSS)
        return []
    endif

    let result = []

    " 当前作用域内的 tagScopes
    let tagScope = omnicpp#complete#GetTagScopeFromScopeStack(a:scopeStack, 1)
    let tagScopes = [tagScope]
    " 添加全局 scope
    if tagScope != '<global>'
        let tagScopes += ['<global>']
    endif

    " 预处理, 主要针对 this-> 和 cast
    let searchScopes = tagScopes[:]
    let memberStack = a:omniSS[:]
    let firstItem = a:omniSS[0]
    if firstItem.kind == 'variable'
        if firstItem.name == 'this'
            " 亦即把 this-> 替换成 MyCls::
            let dNewOmniScope = s:NewOmniScope()
            let dNewOmniScope.name = searchScopes[0]
            let dNewOmniScope.kind = 'container'
            let memberStack = [dNewOmniScope] + a:omniSS[1:]
            " 删除局部作用域, 因为指定 this->, 所以必然从 <global> 搜索此类
            call remove(searchScopes, 0)
        endif
    elseif firstItem.kind == 'cast'
        " TODO: 生成 searchScopes
    endif

    let result = omnicpp#resolvers#DoResolveOmniScopeStack(
                \searchScopes, memberStack, tagScopes)

    return result
endfunc
"}}}

" 递归解析补全请求的补全 scope
" 返回用于获取 tags 的 tagScopes
function! omnicpp#resolvers#DoResolveOmniScopeStack(
            \startTagScopes, memberStack, origTagScopes) "{{{2
    let searchScopes = a:startTagScopes
    let origTagScopes = a:origTagScopes
    " 基本分析方法
    " 1. 解析 member 的实际类型, 并修正 searchScopes
    "    1) 如果是容器, 无须解析
    "    2) 如果是变量, 在作用域中搜索此变量, 然后确定其类型
    "    3) 如果是函数, 在作用域中搜索此函数, 然后确定其返回类型
    " 2. 依次在当前有效 searchScopes 搜索, 从局部到外部, 取首次成功获取的 tag
    " 3. 解析 tag 获取新的 searchScopes, 重复, 直至遍历 memberStack 完毕

    let memberStack = a:memberStack
    let b:memberStack = memberStack
    let memberStackLen = len(memberStack)
    let lVarInfoList = memberStack " 扩展的 memberStack, 
                                   " 会添加 'typeInfo', 'tag' 键
    let prevTag = {}
    let idx = 0
    while idx < memberStackLen
        let member = memberStack[idx]
        " 当前搜索用名称, member 的类型不是 'container', 需要解析
        let curName = member.name

        if member.kind == 'container'
            let curName = member.name
            let typeInfo = s:NewTypeInfo()
            let typeInfo.name = member.name
            let member.typeInfo = typeInfo
        elseif member.kind == 'variable'
            if idx == 0
                " 第一次进入, 因为 ctags 不支持局部变量, 
                " 所以如果起点为变量时, 需要分析是否在当前局部作用域内定义了
                let typeInfo = omnicpp#resolvers#ResolveFirstVariable(
                            \origTagScopes, member.name)
                let tmpTag = s:GetFirstMatchTag(origTagScopes, typeInfo.name)
                let member.typeInfo = typeInfo
                let member.tag = tmpTag
                if !empty(tmpTag)
                    let searchScopes = omnicpp#resolvers#ResolveTag(tmpTag)
                    let prevTag = tmpTag
                    let idx += 1
                    continue
                else
                    " 肯定有错误
                    return []
                endif
            else
                " 常规处理
                let tmpTag = s:GetFirstMatchTag(searchScopes, member.name)

                if has_key(tmpTag, 'typeref')
                    " 变量为无名容器的成员, 直接设置 searchScopes 然后跳过这一步
                    " 因为不会存在无名容器的 path
                    let typerefList = split(tmpTag.typeref, '\w\zs:\ze\w')
                    let kind = typerefList[0]
                    let anon = typerefList[1]
                    let searchScopes = [anon]
                    let idx += 1
                    continue
                endif

                if !empty(tmpTag)
                    let szDecl = tmpTag.cmd
                    let typeInfo = s:GetVariableTypeInfo(szDecl[2:-3], 
                                \tmpTag.name)
                    let member.typeInfo = typeInfo
                    let member.tag = tmpTag
                    if idx - 1 >= 0
                        " 在类外补全
                        let curName = omnicpp#resolvers#ResolveTemplate(
                                    \member, memberStack[idx-1])
                    else
                        " 在类内补全, 唯一的不一样是 tsl 为空
                        let dTmpParentVarType = s:NewOmniScope()
                        let dTmpParentVarType.name = searchScopes[0]
                        let dTmpParentVarType.kind = 'container'
                        let dTmpTypeInfo = s:NewTypeInfo()
                        let dTmpTypeInfo.name = dTmpParentVarType.name
                        let dTmpParentVarType.typeInfo = dTmpTypeInfo
                        let dTmpParentVarType.tag = g:GetTagsByPath(
                                    \dTmpParentVarType.name)
                        let curName = omnicpp#resolvers#ResolveTemplate(
                                    \member, dTmpParentVarType)
                    endif

                    let searchScopes = origTagScopes
                endif
            endif
        elseif member.kind == 'function'
            let tmpTag = s:GetFirstMatchTag(searchScopes, member.name)
            if !empty(tmpTag)
                let szVarType = 
                            \s:GetVariableTypeFromQualifiers(tmpTag.qualifiers)
                if szVarType != ''
                    let typeInfo = s:NewTypeInfo()
                    let typeInfo.name = szVarType
                    let member.typeInfo = typeInfo
                    let member.tag = tmpTag
                    if idx - 1 >= 0
                        " 在类外补全
                        let curName = omnicpp#resolvers#ResolveTemplate(
                                    \member, memberStack[idx-1])
                    else
                        " 在类内补全
                        " TODO
                        let curName = typeInfo.name
                    endif

                    let searchScopes = origTagScopes
                endif
            endif
        endif

        "=======================================================================

        let curTag = s:GetFirstMatchTag(searchScopes, curName)
        let member.tag = curTag
        let searchScopes = omnicpp#resolvers#ResolveTag(curTag)

        let prevTag = curTag
        let idx += 1
    endwhile

    return searchScopes
endfunc
"}}}
" 从限定词中获取模版声明列表
function! s:GetTemplateDeclList(sQualifiers) "{{{2
    let text = a:sQualifiers
    if text == ''
        return []
    endif

    let text = substitute(text, '^.\+<', '', 'g')
    let text = substitute(text, '>.*$', '', 'g')
    let text = substitute(text, 'typename', '', 'g')
    let text = substitute(text, 'class', '', 'g')

    return map(split(text, ','), 's:StripString(v:val)')
endfunc
"}}}
" Param: sSpecDecl 为特化声明. eg. A<B, C<D>, E> Z;
" Return:   模版初始化列表
function! s:GetTemplateSpecList(sSpecDecl) "{{{2
    let s = a:sSpecDecl
    let til = []

    let idx = 0
    let nCount = 0
    let s1 = 0
    while idx < len(s)
        let c = s[idx]
        if c == '<'
            let nCount += 1
            if nCount == 1
                let s1 = idx + 1
            endif
        elseif c == '>'
            let nCount -= 1
            if nCount == 0
                call add(til, s:StripString(s[s1 : idx-1]))
                break
            endif
        elseif c == ','
            if nCount == 1
                call add(til, s:StripString(s[s1 : idx-1]))
                let s1 = idx + 1
            endif
        endif

        let idx += 1
    endwhile

    return til
endfunc
"}}}
" Return: [{'name': 'A', 'tsl': ['B', 'C<D>']}, {'name': 'B', 'tsl': []}]
function! s:GetInheritsInfoList(sInherits) "{{{2
    let sInherits = a:sInherits
    if sInherits == ''
        return []
    endif

    let lResult = []

    let idx = 0
    let nNestDepth = 0
    let nLen = len(sInherits)
    let nSubStart = 0
    while idx < nLen
        let c = sInherits[idx]
        if c == '<'
            let nNestDepth += 1
        elseif c == '>'
            let nNestDepth -= 1
        elseif c == ','
            if nNestDepth == 0
                let sText = sInherits[nSubStart : idx - 1]
                let dInfo = {}
                let dInfo.name = matchstr(sText, '^\s*\zs\w\+\ze\s*<\?')
                let dInfo.tsl = s:GetTemplateSpecList(sText)
                call add(lResult, dInfo)
                let nSubStart = idx + 1
            endif
        endif

        let idx += 1
    endwhile

    let sText = sInherits[nSubStart : idx - 1]
    let dInfo = {}
    let dInfo.name = matchstr(sText, '^\s*\zs\w\+\ze\s*<\?')
    let dInfo.tsl = s:GetTemplateSpecList(sText)
    call add(lResult, dInfo)
    let nSubStart = idx + 1

    return lResult
endfunc
"}}}
" 从父亲的类型的 tag 和 当前变量的模版特化列表, 获取特化到类型声明时的映射
" eg.
" template<class A, class B> class C {A a;};
" C<X, Y> c; c.a.|
" 结果: ['A': 'X', 'B': 'Y']
function! omnicpp#resolvers#GetTemplateMappng(parentTag, tsl) "{{{2
    let parentTag = a:parentTag
    let tsl = a:tsl
    let mapping = {}
    if empty(tsl) || empty(parentTag)
        return {}
    endif

    if parentTag.kind != 'c'
        return {}
    endif

    let tdl = s:GetTemplateDeclList(parentTag.qualifiers)
    if empty(tdl)
        return {}
    endif

    let idx = 0
    while idx < len(tsl)
        let mapping[s:StripString(tdl[idx])] = tsl[idx]
        let idx += 1
    endwhile

    return mapping
endfunc
"}}}
" 在数个 tagScope 中获取第一个匹配的 tag
function! s:GetFirstMatchTag(tagScopes, name) "{{{2
    if a:name == ''
        return {}
    endif

    let tag = {}
    for szScope in a:tagScopes
        let szPath = s:BuildPath(szScope, a:name)
        let tags = g:GetTagsByPath(szPath)
        if !empty(tags)
            let tag = tags[0]
            break
        endif
    endfor
    return tag
endfunc
"}}}
" 从声明中获取变量的类型
" Return: typeInfo
function! s:GetVariableTypeInfo(szDecl, szVar) "{{{2
    let bak_ic = &ic
    set noic

    let typeInfo = s:NewTypeInfo()
    " FIXME: python 的字典转为 vim 字典时, \t 全部丢失
    let szDecl = substitute(a:szDecl, '\\t', ' ', 'g')
    let s = szDecl[: match(szDecl, a:szVar)-1]
    " 删除赋值语句
    let s = substitute(s, '=.*$', '', 'g')
    " eg. int *a, *b, *| -> int |
    let s = substitute(s, '\%(>\|\w\)\zs\s\+\%(\*\+\|&\)\?\s*\w\+\s*,.*$', '', 
                \'g')
    if s =~# '>\s*$'
        " 有模版
        let typeInfo.name = matchstr(s, '\w\+\ze\s*<.\+$')

        " 提取模版初始化列表
        let tsl = matchstr(s, '\w\+\s*\zs<.\+$')
        let typeInfo.tsl = s:GetTemplateSpecList(tsl)
    else
        let typeInfo.name = matchstr(s, '\w\+\ze\s*[*&]\?\s*$')
    endif

    let &ic = bak_ic
    return typeInfo
endfunc
"}}}
" 从限定词中解析变量类型
function! s:GetVariableTypeFromQualifiers(szQualifiers) "{{{2
    let szQualifiers = a:szQualifiers
    if szQualifiers == ''
        return ''
    endif

    " TODO: 解析模版
    return matchstr(szQualifiers, '\w\+\ze\s*\%\(\**\|&\)$')
endfunc
"}}}
" 剔除配对的字符里面(包括配对的字符)的内容
" 可选参数为深度, 从 1 开始, 默认为 1
function! s:StripPair(szString, szStart, szEnd, ...) "{{{2
    let szString = a:szString
    let szStart = a:szStart
    let szEnd = a:szEnd
    let nDeep = 1
    if a:0 > 0
        let nDeep = a:1
    endif

    if nDeep <= 0
        return szString
    endif

    let result = ''

    let idx = 0
    let nLen = len(szString)
    let nCount = 0
    let nSubStart = 0
    while idx < nLen
        let c = szString[idx]
        if c == szStart
            let nCount += 1
            if nCount == nDeep && nSubStart != -1
                let result .= szString[nSubStart : idx - 1]
                let nSubStart = -1
            endif
        elseif c == szEnd
            let nCount -= 1
            if nCount == nDeep - 1
                let nSubStart = idx + 1
            endif
        endif
        let idx += 1
    endwhile

    if nSubStart != -1
        let result .= szString[nSubStart : -1]
    endif

    return result
endfunc
"}}}
" 解析 tag 的 typeref 和 inherits
" 返回与 tag.path 同作用域的 scope 列表
function! omnicpp#resolvers#ResolveTag(tag) "{{{2
    let tagScopes = []
    let tag = a:tag

    if empty(tag)
        return tagScopes
    endif

    let inherits = []

    " 这两个属性不会同时存在?! 'typeref' 比 'inherits' 优先?!
    while has_key(tag, 'typeref')
        let typerefList = split(tag.typeref, '\w\zs:\ze\w')
        let kind = typerefList[0]
        let path = typerefList[1]
        " TODO: 如果是无名容器(eg. MyNs::__anon1), 直接添加到 tagScopes,
        " 因为不会存在无名容器的 path
        let tags = g:GetTagsByKindAndPath(kind, path)
        if empty(tags)
            break
        else
            let tag = tags[0]
        endif
    endwhile

    let tagScopes += [tag.path]

    if has_key(tag, 'inherits')
        let inherits += split(s:StripPair(tag.inherits, '<', '>'), ',')
    endif

    for parent in inherits
        " TODO: 处理复杂的父类作用域
        " TODO: 处理模版类的继承. eg. class A : B<C, D>, E<F> -> B<C, D>,E<F>
        let path = s:BuildPath(tag.scope, parent)

        let tags = g:GetTagsByKindAndPath('c', path)
        if !empty(tags)
            let tagScopes += omnicpp#resolvers#ResolveTag(tags[0])
        endif
    endfor

    return tagScopes
endfunc
"}}}
" 在指定的 searchScopes 中解析补全中的第一个变量的具体类型
" Return: typeInfo ('tag' 键值未填)
function! omnicpp#resolvers#ResolveFirstVariable(searchScopes, szVariable) "{{{2
    " 1. 首先在局部作用域中查找声明
    " 2. 如果查找成功, 直接返回, 否则再在 searchScopes 中查找声明
    let typeInfo = omnicpp#resolvers#SearchLocalDecl(a:szVariable)
    let szVarType = typeInfo.name
    if szVarType == ''
        " 没有在局部作用域到找到此变量的声明
        " 在作用域栈中搜索
        let tmpTag = s:GetFirstMatchTag(a:searchScopes, a:szVariable)
        if !empty(tmpTag)
            let szDecl = tmpTag.cmd
            let typeInfo = s:GetVariableTypeInfo(szDecl[2:-3], tmpTag.name)
        endif
    endif

    return typeInfo
endfunc
"}}}
" 解析模版
" Return: 解析完成的 TB (typename) 的具体值(特化后的类型名)
function! omnicpp#resolvers#ResolveTemplate(dCurVarInfo, dParentVarInfo) "{{{2
    " 同时处理类内的模版补全
    " 如果在类内要求补全模版, 因为没有 tsl, 所以需要检查 tsl
    let dCurVarInfo = a:dCurVarInfo
    let dParentVarInfo = a:dParentVarInfo

    if !has_key(dCurVarInfo.tag, 'class')
        " 不是类成员
        return dCurVarInfo.typeInfo.name
    endif

    let sSpecTypeName = dCurVarInfo.typeInfo.name

    let dReturn = omnicpp#resolvers#DoResolveTemplate(dParentVarInfo.tag, 
                \dCurVarInfo.tag.parent, dCurVarInfo.typeInfo.name)
    if !empty(dReturn)
        if dReturn.flag
            " 关联特化列表
            if !empty(dParentVarInfo.typeInfo.tsl)
                let sSpecTypeName = dParentVarInfo.typeInfo.tsl[dReturn.index]
            else
                " 没有特化列表, 如在类中补全
            endif
        else
            let sSpecTypeName = dReturn.result
        endif
    endif

if 0 "{{{=======================================================================
    let lParentTdl = s:GetTemplateDeclList(dParentVarInfo.tag.qualifiers)

    if dCurVarInfo.tag.parent == dParentVarInfo.typeInfo.name
        let idx = index(lParentTdl, dCurVarInfo.typeInfo.name)
        if idx != -1 && !empty(dParentVarInfo.typeInfo.tsl)
            let sSpecTypeName = dParentVarInfo.typeInfo.tsl[idx]
        endif
    else
        " 当前变量是父类的基类的成员
        " TODO: 现在仅处理了一层继承
        let lInheritsInfoList = s:GetInheritsInfoList(
                    \dParentVarInfo.tag.inherits)
        let bFound = 0
        let tsl = []
        for dInheritsInfo in lInheritsInfoList
            if dInheritsInfo.name == dCurVarInfo.tag.parent
                let bFound = 1
                let tsl = dInheritsInfo.tsl
            endif
        endfor

        let sSubSpecTypeName = ''
        if bFound && !empty(tsl)
            " 需要 tdl
            let dBaseTag = g:GetTagsByPath(dCurVarInfo.tag.scope)[0]
            let tdl = s:GetTemplateDeclList(dBaseTag.qualifiers)
            let idx = index(tdl, dCurVarInfo.typeInfo.name)
            if idx != -1
                let sSubSpecTypeName = tsl[idx]
            endif
        endif

        " 再检查已经解析的特化类型名是不是由父类模版映射的
        let idx = index(lParentTdl, sSubSpecTypeName)
        if idx != -1
            if !empty(dParentVarInfo.typeInfo.tsl)
                let sSpecTypeName = dParentVarInfo.typeInfo.tsl[idx]
            else
                " 类内补全, 并且变量的类型信息为父类模版之一
                " 因为没有特化, 所以无视
            endif
        else
            let sSpecTypeName = sSubSpecTypeName
        endif
    endif
endif "}}}======================================================================

    let b:sSpecTypeName = sSpecTypeName
    return sSpecTypeName
endfunc
"}}}
" 递归解析继承树的模版
" Return: 字典, 返回空表示没有找到
" {
" 'flag': <0|1, 指示是否与派生类的模版相关>, 
" 'index': <模版参数索引, 只有 flag 非 0 时才有用, 0 开始>
" 'result': <解析后的类型名>
" }
function! omnicpp#resolvers#DoResolveTemplate(
            \dCurClassTag, sMatchClass, sTypeName) "{{{2
    let dCurClassTag = a:dCurClassTag
    let sMatchClass = a:sMatchClass
    let sTypeName = a:sTypeName

    let dResult = {'flag': 0, 'index': -1, 'result': sTypeName}

    if dCurClassTag.name == sMatchClass
        let tdl = s:GetTemplateDeclList(dCurClassTag.qualifiers)
        let idx = index(tdl, sTypeName)
        if idx != -1
            let dResult.flag = 1
            let dResult.index = idx
            let dResult.result = tdl[idx]
        else
            let dResult.flag = 0
            let dResult.index = -1
            let dResult.result = sTypeName
        endif
    else
        if !has_key(dCurClassTag, 'inherits')
            " 查找失败
            let dResult = {}
        else
            let sInherits = dCurClassTag.inherits
            let lInheritsInfoList = 
                        \s:GetInheritsInfoList(dCurClassTag.inherits)
            " 深度优先搜索
            for dInheritsInfo in lInheritsInfoList
                let dTags = g:GetTagsByPath(dInheritsInfo.name)
                if empty(dTags)
                    " 应该有语法错误
                    let dResult = {}
                    break
                endif
                let dInheritsTag = dTags[0]
                let dReturn = omnicpp#resolvers#DoResolveTemplate(
                            \dInheritsTag, sMatchClass, sTypeName)
                if !empty(dReturn)
                    " 已经找到
                    if dReturn.flag
                        " 关联当前的特化列表
                        " TODO: 处理语法错误, 即 tsl 为空
                        " 转为当前的类型, 此类型需进一步解析
                        let sTsnTypeName = dInheritsInfo.tsl[dReturn.index]
                        let tdl = s:GetTemplateDeclList(dCurClassTag.qualifiers)
                        let idx = index(tdl, sTsnTypeName)
                        if idx != -1
                            " 关联上层的特化列表
                            let dResult.flag = 1
                            let dResult.index = idx
                            let dResult.result = tdl[idx]
                        else
                            " 确定的类型
                            let dResult.flag = 0
                            let dResult.index = -1
                            let dResult.result = sTsnTypeName
                        endif
                    else
                        " 不关联特化列表, 即确定的类型
                        let dResult.flag = 0
                        let dResult.index = -1
                        let dResult.result = dReturn.result
                    endif

                    break
                endif
            endfor
        endif
    endif

    return dResult
endfunc
"}}}
" 比较两个光标位置 
function! s:CmpPos(pos1, pos2) "{{{2
    let pos1 = a:pos1
    let pos2 = a:pos2
    if pos1[0] > pos2[0]
        return 1
    elseif pos1[0] < pos2[0]
        return -1
    else
        if pos1[1] > pos2[1]
            return 1
        elseif pos1[1] < pos2[1]
            return -1
        else
            return 0
        endif
    endif
endfunc
"}}}
" 获取向上搜索局部变量的停止位置, 同时是也是向下搜索的开始位置
" Note: 如果光标不在任何 '局部' 作用域中, 即为全局作用域, 返回 [0, 0]
"       全局作用域不应调用此函数
" eg.
" class A {
"     char foo;
"     class B {
"         void F()
"         {
"             long foo;
"         }
"         short foo;
"         void Func(int foo)
"         {
"             foo = 0;|
"         }
"     };
" };
" TODO: 处理 extern "C" {
function! omnicpp#resolvers#GetStopPosForLocalDeclSearch(...) "{{{2
    let origCursor = getpos('.')
    let origPos = origCursor[1:2]
    let stopPos = origPos
    let curPos = origPos

    let scopeStack = omnicpp#scopes#GetScopeStack()
    " 从作用域栈顶部开始, 一直向上搜索块开始位置
    " 直至遇到非 'other' 类型的作用域,
    for scope in reverse(scopeStack)
        let curBlkStaPos = omnicpp#scopes#GetCurBlockStartPos(1)
        if curBlkStaPos == [0, 0]
            " 到达这里, 肯定有语法错误, 或者是全局作用域
            let stopPos = [0, 0]
            break
        endif

        if scope.kind != 'other'
            break
        endif
    endfor

    if stopPos != [0, 0]
        let stopPos = omnicpp#utils#GetCurStatementStartPos(1)
    endif

    if a:0 > 0 && a:1
    else
        call setpos('.', origCursor)
    endif

    return stopPos
endfunc
"}}}
" Search a local declaration
" Return: typeInfo
" 搜索局部变量的声明类型. 
" TODO: 这个函数很慢! 需要优化. 需要变量分析的场合的速度比不需要的场合的慢 3 倍!
function! omnicpp#resolvers#SearchLocalDecl(szVariable) "{{{2
    let typeInfo = s:NewTypeInfo()
    let origCursor = getpos('.')
    let origPos = origCursor[1:2]

    let bak_ic = &ignorecase
    set noignorecase

    " TODO: 这个函数比较慢
    let stopPos = omnicpp#resolvers#GetStopPosForLocalDeclSearch(1)
    let curPos = stopPos[:]
    while curPos != [0, 0]
        " 1. 往下搜索不超过起始位置的同名符号, 解析 tokens
        " 2. 若为无效的声明, 重复, 否则结束
        " " TODO: 是否应该反转方向搜索? 因为现在是外层变量声明覆盖了内层
        let curPos = searchpos('\C\<'. a:szVariable .'\>', 'W')

        " TODO: 这两个函数很慢
        " 原因是, 前面或者后面嵌套的大括号太多, 想办法优化
        if 0
            if s:CmpPos(omnicpp#scopes#GetCurBlockEndPos(), origPos) < 0
                " 肯定进入了其他的 {} 块, 或者在函数参数中
                if s:CmpPos(omnicpp#scopes#GetCurBlockStartPos(), stopPos) > 0
                    " 肯定进入了其他的 {} 块
                    continue
                endif
            endif
        endif

        if curPos != [0, 0] && s:CmpPos(curPos, origPos) < 0
            " TODO: 验证是否有效的声明
            " 搜索 foo 声明行应该为 Foo foo;
            " TestCase:
            " mc.foo();
            " Foo foo;
            " foo;

            let tokens = omnicpp#utils#TokenizeStatementBeforeCursor()

            " 上述情形, tokens 最后项的类型只能为
            " C++ 关键词或单词
            " * 或 & 后 > 操作符(eg. void *p; int &n; vector<int> x;)
            " TODO: 使用 tokens 检查
            if !empty(tokens)
                if tokens[-1].kind == 'cppOperatorPunctuator' 
                            \&& tokens[-1].value =~# '\V.\|->'
                    " 无效的声明, 继续
                    continue
                endif
            endif

            " 解析声明
            " FIXME: 如果为原始类型, 如 int, short, 返回空字符串!
            let typeInfo.name = omnicpp#utils#GetVariableType(tokens)

            let szCurStmt = omnicpp#utils#GetCurStatementBeforeCursor()
            let typeInfo.tsl = s:GetTemplateSpecList(szCurStmt)
            break
        else
            " 搜索失败
            break
        endif
    endwhile

    call setpos('.', origCursor)
    let &ignorecase = bak_ic

    return typeInfo
endfunc
"}}}
" vim:fdm=marker:fen:expandtab:smarttab:fdl=1:
autoload/omnicpp/VimTagsManager.py	[[[1
104
#!/usr/bin/env python
# -*- coding:utf-8 -*-

import TagsStorageSQLite as TagsStorage
from TagEntry import ToFullKind, ToFullKinds



def TagEntry2Tag(tagEntry):
    '''把 python 的 TagEntry 转为 vim 的 tag 字典'''
    tag = {} # taglist() 返回的列表的项目
    # 必不可少的五个字段
    # omnicpp 需要完整的名字
    #tag['name'] = tagEntry.GetPath().encode('utf-8')
    tag['name'] = tagEntry.GetName().encode('utf-8')
    tag['filename'] = tagEntry.GetFile().encode('utf-8')
    tag['cmd'] = tagEntry.GetPattern().encode('utf-8')
    # 全称改为简称, 用命令参数控制
    tag['kind'] = tagEntry.GetAbbrKind().encode('utf-8')
    tag['static'] = 0 # 作用不明

    # 附加
    tag['parent'] = tagEntry.GetParent().encode('utf-8')
    tag['scope'] = tagEntry.GetScope().encode('utf-8')
    tag['path'] = tagEntry.GetPath().encode('utf-8')
    tag['qualifiers'] = tagEntry.GetQualifiers().encode('utf-8')

    # 附加字段
    if tagEntry.GetAccess():
        tag['access'] = tagEntry.GetAccess().encode('utf-8')
    if tagEntry.GetInherits():
        tag['inherits'] = tagEntry.GetInherits().encode('utf-8')
    if tagEntry.GetSignature():
        tag['signature'] = tagEntry.GetSignature().encode('utf-8')
    if tagEntry.GetTyperef():
        tag['typeref'] = tagEntry.GetTyperef().encode('utf-8')
    if tagEntry.GetParentType():
        tag[tagEntry.GetParentType().encode('utf-8')]\
                = tagEntry.GetParent().encode('utf-8')
    # FIXME: class:MyClass, struct:MyStruce, union:MyUnion, 无能为力
    #if tagEntry.

    return tag

def TagEntries2Tags(tagEntries):
    tags = [] # 符合 vim 接口的 tags 列表
    tags = []
    for tagEntry in tagEntries:
        tags.append(TagEntry2Tag(tagEntry))
    return tags



class VimTagsManager:
    def __init__(self, dbFile = ''):
        self.storage = TagsStorage.TagsStorageSQLite()
        if dbFile:
            self.storage.OpenDatabase(dbFile)

    def OpenDatabase(self, dbFile):
        self.storage.OpenDatabase(dbFile)

    def ParseFiles(self, files):
        TagsStorage.ParseFilesAndStore(files, self.storage)

    def GetTagsByScopeAndName(self, scope, name):
        tagEntries = self.storage.GetTagsByScopeAndName(scope, name, True)
        return TagEntries2Tags(tagEntries)

    def GetTagsByScopesAndName(self, scopes, name):
        tagEntries = self.storage.GetTagsByScopeAndName(scopes, name, True)
        return TagEntries2Tags(tagEntries)

    def GetTagsByScopeAndKind(self, scope, kind):
        return self.GetTagsByScopesAndKinds([scope], [kind])

    def GetTagsByScopesAndKinds(self, scopes, kinds):
        tagEntries = self.storage.GetTagsByScopesAndKinds(scopes, 
                                                          ToFullKinds(kinds))
        return TagEntries2Tags(tagEntries)

    def GetTagsByPath(self, path):
        tagEntries = self.storage.GetTagsByPath(path)
        return TagEntries2Tags(tagEntries)

    def GetTagsByKindAndPath(self, kind, path):
        tagEntries = self.storage.GetTagsByKindAndPath(ToFullKind(kind), path)
        return TagEntries2Tags(tagEntries)

if __name__ == '__main__':
    vtm = VimTagsManager()
    vtm.OpenDatabase('TestTags3.db')
    print vtm.storage.fileName
    #tags = vtm.GetTagsByScopesAndKinds(
        #['DFoo', 'Foo', 'Foo2'], ['member', 'function'])
    #print str(tags)
    #tags = vtm.GetTagsByScopesAndName(['DFoo', 'Foo', 'Foo2'], '')
    #tags = vtm.GetTagsByScopeAndName(['<global>'], 'test')
    tags = vtm.GetTagsByPath('TestNamespace')
    print tags
    print len(tags)
    for tag in tags:
        print tag['name']

autoload/omnicpp/TagsStorageSQLite.py	[[[1
1275
#!/usr/bin/env python
# -*- encoding:utf-8 -*-


from ITagsStorage import ITagsStorage
from TagEntry import TagEntry
from FileEntry import FileEntry

import os, os.path
import sqlite3

class TagsStorageSQLiteCache:
    '''tags 缓存'''
    def __init__(self):
        # 以 sql 语句为键值
        self.cache = {} # 字符串到标签条目列表的字典 {'fo': ['foo', 'foobar']}

    def DoGet(self, key):
        tags = []
        if self.cache.has_key(key):
            tags = self.cache[key][:]
        return tags

    def DoStore(self, key, tags):
        '''以 sql 语句为键值保存 tags 的副本'''
        self.cache[key] = tags[:]

    def Get(self, sql, kinds = []):
        if not kinds:
            return self.DoGet(sql)
        else:
            key = sql
            for kind in kinds:
                key += "@" + kind
            return self.DoGet(key)

    def Store(self, sql, tags, kinds = []):
        if not kinds:
            return self.DoStore(sql, tags)
        else:
            key = sql
            for kind in kinds:
                # 不是原始的 sql 语句, 是简化过的
                key += "@" + kind
            return self.DoStore(key, tags)

    def Clear(self):
        self.cache.clear()

tagsDatabaseVersion = 'VimLite Version 1.0'

class TagsStorageSQLite(ITagsStorage):
    def __init__(self):
        ITagsStorage.__init__(self)
        self.db = None      # sqlite3 的连接实例, 取此名字是为了与 codelite 统一
        self.cache = None   # TagsStorageSQLiteCache 类的实例

        # 测试时关闭缓存
        self.SetUseCache(False)

    def __del__(self):
        if self.db:
            self.db.close()
            self.db = None

    def GetVersion(self):
        return tagsDatabaseVersion

    def Begin(self):
        if self.db:
            try:
                self.db.execute("begin;")
            except sqlite3.OperationalError:
                pass

    def Commit(self):
        if self.db:
            try:
                self.db.commit()
            except sqlite3.OperationalError:
                pass

    def OpenDatabase(self, fileName = ''):
        # TODO: 验证文件是否有效

        # 如果相同, 表示已经打开了相同的数据库, 直接返回
        if self.fileName == os.path.abspath(fileName):
            return True

        # Did we get a file name to use?
        # 未打开任何数据库, 且请求打开的文件无效, 直接返回
        if not self.fileName and not fileName:
            return False

        # We did not get any file name to use BUT we
        # do have an open database, so we will use it
        # 传进来的是无效的文件, 但已经打开了某个数据库, 继续用之
        if not fileName:
            return True

        try:
            if not self.fileName:
                # First time we open the db
                # 没有打开着的数据库
                self.db = sqlite3.connect(os.path.abspath(fileName))
                self.CreateSchema()
                self.fileName = os.path.abspath(fileName)
            else:
                # We have both fileName & self.fileName and they
                # are different, close previous db
                # 已经打开了某个数据库, 请求打开另外的, 需要先关闭旧的
                self.db.close()
                self.db = sqlite3.connect(os.path.abspath(fileName))
                self.CreateSchema()
                self.fileName = os.path.abspath(fileName)
            return True
        except sqlite3.OperationalError:
            return False

    def CreateSchema(self):
        try:
            # improve performace by using pragma command:
            # (this needs to be done before the creation of the
            # tables and indices)
            # Is it valid in pysqlite3?
            sql = "PRAGMA synchronous = OFF;"
            self.db.execute(sql)

            sql = "PRAGMA temp_store = MEMORY;"
            self.db.execute(sql)

            # 添加 parentType 以用于后向兼容
            sql = "create  table if not exists tags (ID INTEGER PRIMARY KEY "\
                    "AUTOINCREMENT, name string, file string, line integer, "\
                    "kind string, access string, signature string, pattern "\
                    "string, parent string, inherits string, path string, "\
                    "typeref string, scope string, return_value string, "\
                    "parent_type string, qualifiers string);"
            self.db.execute(sql)

            sql = "create  table if not exists FILES (ID INTEGER PRIMARY KEY "\
                    "AUTOINCREMENT, file string, last_retagged integer);"
            self.db.execute(sql)

            sql = "create  table if not exists MACROS (ID INTEGER PRIMARY KEY "\
                    "AUTOINCREMENT, file string, line integer, name string, "\
                    "is_function_like int, replacement string, signature "\
                    "string);"
            self.db.execute(sql)

            sql = "create  table if not exists SIMPLE_MACROS (ID INTEGER "\
                    "PRIMARY KEY AUTOINCREMENT, file string, name string);"
            self.db.execute(sql)

            # create unuque index on Files' file column
            sql = "CREATE UNIQUE INDEX IF NOT EXISTS FILES_NAME on FILES(file)"
            self.db.execute(sql)

            # Create unique index on tags table
            sql = "CREATE UNIQUE INDEX IF NOT EXISTS TAGS_UNIQ on tags(kind, "\
                    "path, signature);"
            self.db.execute(sql)

            sql = "CREATE INDEX IF NOT EXISTS KIND_IDX on tags(kind);"
            self.db.execute(sql)

            sql = "CREATE INDEX IF NOT EXISTS FILE_IDX on tags(file);"
            self.db.execute(sql)

            sql = "CREATE UNIQUE INDEX IF NOT EXISTS MACROS_UNIQ on "\
                    "MACROS(name);"
            self.db.execute(sql)

            # Create search indexes
            sql = "CREATE INDEX IF NOT EXISTS TAGS_NAME on tags(name);"
            self.db.execute(sql)

            sql = "CREATE INDEX IF NOT EXISTS TAGS_SCOPE on tags(scope);"
            self.db.execute(sql)

            sql = "CREATE INDEX IF NOT EXISTS TAGS_PATH on tags(path);"
            self.db.execute(sql)

            sql = "CREATE INDEX IF NOT EXISTS TAGS_PARENT on tags(parent);"
            self.db.execute(sql)

            sql = "CREATE INDEX IF NOT EXISTS MACROS_NAME on MACROS(name);"
            self.db.execute(sql)

            sql = "CREATE INDEX IF NOT EXISTS SIMPLE_MACROS_FILE on "\
                    "SIMPLE_MACROS(file);"
            self.db.execute(sql)

            sql = "create table if not exists tags_version (version string "\
                    "primary key);"
            self.db.execute(sql)

            sql = "create unique index if not exists tags_version_uniq on "\
                    "tags_version(version);"
            self.db.execute(sql)

            #sql = "insert into tags_version values ('"+ self.GetVersion() +"');"
            # TODO: 没有替换
            sql = "insert or replace into tags_version values ('" \
                    + self.GetVersion() + "');"
            self.db.execute(sql)

            # 必须提交
            self.db.commit()
        except sqlite3.OperationalError:
            pass

    def RecreateDatabase(self):
        if not self.fileName:
            return

        try:
            # commit any open transactions
            self.Commit()

            # Close the database
            if self.db:
                self.db.close()

            try:
                # TODO
                os.remove(self.fileName)
            except:
                # re-open the database
                self.OpenDatabase(self.fileName)

                # and drop tables
                self.db.execute("DROP TABLE IF EXISTS TAGS")
                self.db.execute("DROP TABLE IF EXISTS COMMENTS")
                self.db.execute("DROP TABLE IF EXISTS TAGS_VERSION")
                self.db.execute("DROP TABLE IF EXISTS VARIABLES")
                self.db.execute("DROP TABLE IF EXISTS FILES")
                self.db.execute("DROP TABLE IF EXISTS MACROS")
                self.db.execute("DROP TABLE IF EXISTS SIMPLE_MACROS")

                # drop indexes
                self.db.execute("DROP INDEX IF EXISTS FILES_NAME")
                self.db.execute("DROP INDEX IF EXISTS TAGS_UNIQ")
                self.db.execute("DROP INDEX IF EXISTS KIND_IDX")
                self.db.execute("DROP INDEX IF EXISTS FILE_IDX")
                self.db.execute("DROP INDEX IF EXISTS TAGS_NAME")
                self.db.execute("DROP INDEX IF EXISTS TAGS_SCOPE")
                self.db.execute("DROP INDEX IF EXISTS TAGS_PATH")
                self.db.execute("DROP INDEX IF EXISTS TAGS_PARENT")
                self.db.execute("DROP INDEX IF EXISTS tags_version_uniq")
                self.db.execute("DROP INDEX IF EXISTS MACROS_UNIQ")
                self.db.execute("DROP INDEX IF EXISTS MACROS_NAME")
                self.db.execute("DROP INDEX IF EXISTS SIMPLE_MACROS_FILE")

                # Recreate the schema
                self.CreateSchema()
            else:
                # We managed to delete the file
                # re-open it
                self.OpenDatabase(self.fileName)
        except sqlite3.OperationalError:
            pass

    def GetSchemaVersion(self):
        try:
            version = ''
            sql = "SELECT * FROM TAGS_VERSION"
            for row in self.db.execute(sql):
                version = row[0]
            return version
        except sqlite3.OperationalError:
            pass
        return ''

    def Store(self, tagTree, dbFile = '', autoCommit = True):
        '''需要 tagTree
        
        tagTree 为标签的 path 树'''

        # 现阶段, 支持直接从标签文本保存
        if isinstance(tagTree, str):
            # 直接从字符串保存
            tags = tagTree

            if not dbFile and not self.fileName:
                return

            if not tags:
                return

            self.OpenDatabase(dbFile) # 这里, 如果 dbFile 为空, 表示使用原来的
            try:
                updateList = [] # 不存在的直接插入, 存在的需要更新

                if autoCommit:
                    # FIXME: 这样真的大丈夫?
                    self.db.execute('begin;')

                for line in tags.split('\n'):
                    # does not matter if we insert or update, 
                    # the cache must be cleared for any related tags

                    tagEntry = TagEntry()
                    tagEntry.FromLine(line)

                    if not self.InsertTagEntry(tagEntry):
                        # 插入不成功?
                        # InsertTagEntry() 貌似是不会失败的?!
                        updateList.append(tagEntry)

                if autoCommit:
                    self.db.commit()

                # Do we need to update?
                if updateList:
                    if autoCommit:
                        # FIXME: 这样真的大丈夫?
                        self.db.execute('begin;')

                    for i in updateList:
                        self.UpdateTagEntry(i)

                    if autoCommit:
                        self.db.commit()
            except sqlite3.OperationalError:
                try:
                    if autoCommit:
                        self.db.rollback()
                except sqlite3.OperationalError:
                    pass
        else:
            pass
        pass

    def SelectTagsByFile(self, file, dbFile = ''):
        '''取出属于 file 文件的全部标签'''
        # Incase empty dbFile is provided, use the current file name
        if not dbFile:
            dbFile = self.fileName
        self.OpenDatabase(dbFile)

        sql = "select * from tags where file='" + file + "' "
        return self.DoFetchTags(sql)

    def DeleteByFileName(self, dbFile, fileName, autoCommit = True):
        '''删除属于指定文件名 fileName 的所有标签'''
        # make sure database is open
        try:
            self.OpenDatabase(dbFile)
            if autoCommit:
                # FIXME: 这样真的大丈夫?
                self.db.execute('begin;')

            # TODO: 不安全?
            self.db.execute("Delete from tags where File='%s'" % fileName)

            if autoCommit:
                self.db.commit()
        except:
            if autoCommit:
                self.db.rollback()

    def Query(self, sql, dbFile = ''):
        '''Execute a query sql and return result set.
        
        这个函数特别之处在于自动 OpenDatabase()'''
        try:
            self.OpenDatabase(dbFile)
            return self.db.execute(sql)
        except:
            pass
        return [] # 具备迭代器的空对象

    def ExecuteUpdate(self, sql):
        try:
            self.db.execute(sql)
        except:
            pass

    def IsOpen(self):
        if self.db:
            return True
        else:
            return False

    def GetFiles(self, partialName = ''):
        files = []

        if not partialName:
            try:
                sql = "select * from files order by file"
                res = self.db.execute(sql)
                for row in res:
                    fe = FileEntry()
                    fe.SetId(row[0])
                    fe.SetFile(row[1])
                    fe.SetLastRetaggedTimestamp(row[2])
                    files.append(fe)
            except:
                pass
        else:
            try:
                matchPath = partialName and partialName.endswith(os.sep)
                tmpName = partialName.replace('_', '^_')
                sql = "select * from files where file like '%%" + tmpName \
                        + "%%' ESCAPE '^' "
                res = self.db.execute(sql)
                for row in res:
                    fe = FileEntry()
                    fe.SetId(row[0])
                    fe.SetFile(row[1])
                    fe.SetLastRetaggedTimestamp(row[2])

                    fileName = fe.GetFile()
                    match = os.path.basename(fileName)
                    if matchPath:
                        match = fileName

                    # TODO: windows 下文件名全部保存为小写

                    if match.startswith(partialName):
                        files.append(fe)
            except:
                pass

        return files

    def GetFilesMap(self, partialName = ''):
        '''返回文件到文件条目的字典, 方便比较'''
        filesMap = {}

        if not partialName:
            try:
                sql = "select * from files order by file"
                res = self.db.execute(sql)
                for row in res:
                    fe = FileEntry()
                    fe.SetId(row[0])
                    fe.SetFile(row[1])
                    fe.SetLastRetaggedTimestamp(row[2])
                    filesMap[fe.GetFile()] = fe
            except:
                pass
        else:
            try:
                matchPath = partialName and partialName.endswith(os.sep)
                tmpName = partialName.replace('_', '^_')
                sql = "select * from files where file like '%%" + tmpName \
                        + "%%' ESCAPE '^' "
                res = self.db.execute(sql)
                for row in res:
                    fe = FileEntry()
                    fe.SetId(row[0])
                    fe.SetFile(row[1])
                    fe.SetLastRetaggedTimestamp(row[2])

                    fileName = fe.GetFile()
                    match = os.path.basename(fileName)
                    if matchPath:
                        match = fileName

                    # TODO: windows 下文件名全部保存为小写

                    if match.startswith(partialName):
                        filesMap[fe.GetFile()] = fe
            except:
                pass

        return filesMap

    def LastRowId(self):
        # 不需要
        pass

    def DeleteByFilePrefix(self, dbFile, filePrefix):
        try:
            self.OpenDatabase(dbFile)
            sql = "delete from tags where file like '" \
                    + filePrefix.replace('_', '^_') + "%%' ESCAPE '^' "
            self.db.execute(sql)
        except:
            pass

    def DeleteFromFiles(self, files):
        if not files:
            return

        sql = "delete from FILES where file in ("
        for file in files:
            sql += "'" + file + "',"

        # remove last ','
        sql = sql[:-1] + ')'

        try:
            self.db.execute(sql)
        except:
            pass

    def DeleteFromFilesByPrefix(self, dbFile, filePrefix):
        try:
            self.OpenDatabase(dbFile)
            sql = "delete from FILES where file like '" \
                    + filePrefix.replace('_', '^_') + "%%' ESCAPE '^' "
            self.db.execute(sql)
        except:
            pass

    def PPTokenFromSQlite3ResultSet(self, rs, token):
        pass

    def FromSQLite3ResultSet(self, row):
        '''从数据库的一行数据中提取标签对象'''
        # 添加 parentType
        entry = TagEntry()
        entry.SetId         (row[0])
        entry.SetName       (row[1])
        entry.SetFile       (row[2])
        entry.SetLine       (row[3])
        entry.SetKind       (row[4])
        entry.SetAccess     (row[5])
        entry.SetSignature  (row[6])
        entry.SetPattern    (row[7])
        entry.SetParent     (row[8])
        entry.SetInherits   (row[9])
        entry.SetPath       (row[10])
        entry.SetTyperef    (row[11])
        entry.SetScope      (row[12])
        entry.SetReturnValue(row[13])
        entry.SetParentType (row[14])
        entry.SetQualifiers (row[15])

        return entry

    def DoFetchTags(self, sql, kinds = []):
        '''从数据库中取出 tags'''
        tags = []

        if not kinds:
            if self.GetUseCache():
                # 尝试从缓存中获取
                tags = self.cache.Get(sql)
                if tags:
                    print '[CACHED ITEMS] %s\n' % sql
                    return

            try:
                exRs = self.Query(sql)

                # add results from external database to the workspace database
                for row in exRs:
                    tag = self.FromSQLite3ResultSet(row)
                    tags.append(tag)
            except:
                pass

            if self.GetUseCache():
                # 保存到缓存以供下次快速使用
                self.cache.Store(sql, tags)
        else:
            if self.GetUseCache():
                # 尝试从缓存中获取
                tags = self.cache.Get(sql, kinds)
                if tags:
                    print '[CACHED ITEMS] %s\n' % sql
                    return

            try:
                exRs = self.Query(sql)
                for row in exRs:
                    try:
                        kinds.index(row[4])
                    except ValueError:
                        continue
                    else:
                        tag = self.FromSQLite3ResultSet(row)
                        tags.append(tag)
            except:
                pass

            if self.GetUseCache():
                # 保存到缓存以供下次快速使用
                self.cache.Store(sql, tags, kinds)

        return tags

    def GetTagsByScopeAndName(self, scope, name, partialNameAllowed = False):
        if type(scope) == type(''):
            tmpName = name.replace('_', '^_')

            sql = "select * from tags where "

            # did we get scope?
            if scope:
                sql += "scope='" + scope + "' and "

            # add the name condition
            if partialNameAllowed:
                sql += " name like '" + tmpName + "%%' ESCAPE '^' "
            else:
                sql += " name ='" + name + "' "

            sql += " LIMIT " + str(self.GetSingleSearchLimit())

            # get the tags
            return self.DoFetchTags(sql)
        elif type(scope) == type([]):
            scopes = scope
            if not scopes:
                return []
            tmpName = name.replace('_', '^_')

            sql = "select * from tags where scope in("
            for i in scopes:
                sql += "'" + i + "',"
            sql = sql[:-1] + ") and "

            # add the name condition
            if partialNameAllowed:
                sql += " name like '" + tmpName + "%%' ESCAPE '^' "
            else:
                sql += " name ='" + name + "' "

            # get the tags
            return self.DoFetchTags(sql)

        else:
            return []

    def GetTagsByScope(self, scope):
        sql = "select * from tags where scope='" + scope + "' limit " \
                + str(self.GetSingleSearchLimit())
        return self.DoFetchTags(sql)

    def GetTagsByKinds(self, kinds, orderingColumn, order):
        sql = "select * from tags where kind in ("
        for i in kinds:
            sql += "'" + i + "',"

        sql = sql[:-1] + ") "

        if orderingColumn:
            sql += "order by " + orderingColumn
            if order == ITagsStorage.OrderAsc:
                sql += " ASC"
            elif order == ITagsStorage.OrderDesc:
                sql += " DESC"
            else:
                pass

        return self.DoFetchTags(sql)

    def GetTagsByPath(self, path):
        if type(path) == type([]):
            sql = "select * from tags where path IN("
            for i in path:
                sql += "'" + i + "',"
            sql = sql[:-1] + ") "
            return self.DoFetchTags(sql)
        else:
            sql = "select * from tags where path ='" + path + "' LIMIT 1"
            return self.DoFetchTags(sql)

    def GetTagsByPaths(self, paths):
        return self.GetTagsByPaths(paths)

    def GetTagsByNameAndParent(self, name, parent):
        '''根据标签名称和其父亲获取标签'''
        sql = "select * from tags where name='" + name + "'"
        tags = self.DoFetchTags(sql)

        # 过滤掉不符合要求的标签
        return [i for i in tags if i.GetParent() == parent]

    def GetTagsByKindsAndPath(self, kinds, path):
        if not kinds:
            return []

        sql = "select * from tags where path='" + path + "'"
        return self.DoFetchTags(sql, kinds)

    def GetTagsByKindAndPath(self, kind, path):
        return self.GetTagsByKindsAndPath([kind], path)

    def GetTagsByFileAndLine(self, file, line):
        sql = "select * from tags where file='" + file \
                + "' and line=" + line + " "
        return self.DoFetchTags(sql)

    def GetTagsByScopeAndKind(self, scope, kind):
        return self.GetTagsByScopesAndKinds([scope], [kind])

    def GetTagsByScopeAndKinds(self, scope, kinds):
        if not kinds:
            return []

        sql = "select * from tags where scope='" + scope + "'"
        return self.DoFetchTags(sql, kinds)

    def GetTagsByKindsAndFile(self, kinds, fileName, orderingColumn, order):
        if not kinds:
            return []

        sql = "select * from tags where file='" + fileName + "' and kind in ("
        for i in kinds:
            sql += "'" + i + "',"
        sql = sql[:-1] + ")"

        if orderingColumn:
            sql += "order by " + orderingColumn
            if order == ITagsStorage.OrderAsc:
                sql += " ASC"
            elif order == ITagsStorage.OrderDesc:
                sql += " DESC"
            else:
                pass

        return self.DoFetchTags(sql)

    def DeleteFileEntry(self, fileName):
        try:
            self.db.execute("DELETE FROM FILES WHERE FILE=?", (fileName,))
            self.db.commit()
        except:
            # TODO: 区分错误代码
            return False
        else:
            return True

    def InsertFileEntry(self, fileName, timestamp):
        try:
            # 理论上, 不会插入失败
            self.db.execute("INSERT OR REPLACE INTO FILES VALUES(NULL, ?, ?)", 
                           (fileName, timestamp))
            self.db.commit()
        except:
            return False
        else:
            return True

    def UpdateFileEntry(self, fileName, timestamp):
        try:
            self.db.execute(
                "UPDATE OR REPLACE FILES SET last_retagged=? WHERE file=?", 
                (timestamp, fileName))
            self.db.commit()
        except:
            return False
        else:
            return True

    def DeleteTagEntry(self, kind, signature, path):
        try:
            self.db.execute(
                "DELETE FROM TAGS WHERE Kind=? AND Signature=? AND Path=?", 
                (kind, signature, path))
            self.db.commit()
        except:
            return False
        else:
            return True

    def InsertTagEntry(self, tag):
        if not tag.IsOk():
            return True

        if self.GetUseCache():
            self.ClearCache()

        #try:
        if 1:
            # INSERT OR REPLACE 貌似是不会失败的?!
            # FIXME: sql 语句必须是 unicode 编码, 而文件编码一般为 utf-8
            # 需要 decode. 只有模式可能有其他字符?
            # 添加 parentType
            self.db.execute(
                "INSERT OR REPLACE INTO TAGS VALUES (NULL, "\
                "?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", 
                (tag.GetName(), 
                 tag.GetFile(), 
                 tag.GetLine(), 
                 tag.GetKind(), 
                 tag.GetAccess(), 
                 tag.GetSignature(), 
                 tag.GetPattern().decode('utf-8'), 
                 tag.GetParent(), 
                 tag.GetInheritsAsString(), 
                 tag.GetPath(), 
                 tag.GetTyperef(), 
                 tag.GetScope(), 
                 tag.GetReturnValue(), 
                 tag.GetParentType(), 
                 tag.GetQualifiers()))
            self.db.commit()
        #except:
            #return False
        else:
            return True

    def UpdateTagEntry(self, tag):
        if not tag.IsOk():
            return True

        if self.GetUseCache():
            self.ClearCache()

        try:
            # 添加 parentType
            self.db.execute(
                "UPDATE OR REPLACE TAGS SET "
                "Name=?, File=?, Line=?, Access=?, Pattern=?, Parent=?, "
                "Inherits=?, Typeref=?, Scope=?, Return_Value=?, "
                "Parenttype=?, Qualifiers=? "
                "WHERE Kind=? AND Signature=? AND Path=?", 
                (tag.GetName(), 
                 tag.GetFile(), 
                 tag.GetLine(), 
                 tag.GetAccess(), 
                 tag.GetPattern().decode('utf-8'), 
                 tag.GetParent(), 
                 tag.GetInheritsAsString(), 
                 tag.GetTyperef(), 
                 tag.GetScope(), 
                 tag.GetReturnValue(), 
                 tag.GetParenttype(), 
                 tag.GetQualifiers(), 

                # where ? 这三个参数能唯一定位?
                 tag.GetKind(), 
                 tag.GetSignature(), 
                 tag.GetPath()))
            self.db.commit()
        except:
            return False
        else:
            return True

    def IsTypeAndScopeContainer(self, typeName, scope):
        '''返回有三个元素的元组 (Ture/False, typeName, scope)
        True if type exist under a given scope.
        
        Incase it exist but under the <global> scope, 'scope' will be changed
        '''
        # Break the typename to 'name' and scope
        typeNameNoScope = typeName.rpartition(':')[2]
        scopeOne        = typeName.rpartition(':')[0]

        if scopeOne.endswith(':'):
            scopeOne = scopeOne[:-1]

        combinedScope = ''

        if scope != '<global>':
            combinedScope += scope

        if scopeOne:
            if combinedScope:
                combinedScope += '::'
            combinedScope += scopeOne

        sql = "select scope,kind from tags where name='" + typeNameNoScope + "'"

        foundGlobal = False

        try:
            for row in self.Query(sql):
                scopeFounded = row[0]
                kindFounded = row[1]
                containerKind = kindFounded == "struct" \
                        or kindFounded == "class"
                if scopeFounded == combinedScope and containerKind:
                    scope = combinedScope
                    typeName = typeNameNoScope
                    return True, typeName, scope
                elif scopeFounded == scopeOne and containerKind:
                    # this is equal to cases like this:
                    # class A {
                    #     typedef std::list<int> List;
                    #     List l;
                    # };
                    # the combinedScope will be: 'A::std'
                    # however, the actual scope is 'std'
                    scope = scopeOne
                    typeName = typeNameNoScope
                    return True, typeName, scope
                elif containerKind and scopeFounded == "<global>":
                    foundGlobal = True
        except:
            pass

        if foundGlobal:
            scope = "<global>"
            typeName = typeNameNoScope
            return True, typeName, scope

        return False, typeName, scope

    def IsTypeAndScopeExist(self, typeName, scope):
        bestScope = ''
        tmpScope = scope

        strippedName    = typeName.rpartition(':')[2]
        secondScope     = typeName.rpartition(':')[0]

        if secondScope.endswith(':'):
            secondScope = secondScope[:-1]

        if not strippedName:
            return False

        sql = "select scope,parent from tags where name='" + strippedName \
                + "' and kind in ('class', 'struct', 'typedef') LIMIT 50"
        foundOther = 0

        if secondScope:
            tmpScope += '::' + secondScope

        parent = tmpScope.rpartition(':')[2]

        try:
            for row in self.Query(sql):
                scopeFounded = row[0]
                parentFounded = row[1]

                if scopeFounded == tmpScope:
                    scope = scopeFounded
                    typeName = strippedName
                    return True
                elif parentFounded == parent:
                    bestScope = scopeFounded
                else:
                    foundOther += 1
        except:
            pass

        # if we reached here, it means we did not find any exact match
        if bestScope:
            scope = bestScope
            typeName = strippedName
            return True
        elif foundOther == 1:
            scope = scopeFounded
            typeName = strippedName
            return True

        return False

    def GetScopesFromFileAsc(self, fileName, scopes):
        '''传入的 scopes 为列表'''
        sql = "select * from tags where file = '" + fileName + "' " \
                + " and kind in('prototype', 'function', 'enum')" \
                + " order by scope ASC"

        # we take the first entry
        try:
            for row in self.Query(sql):
                scopes.append(row[0])
                break
        except:
            pass

    def GetTagsByFileScopeAndKinds(self, fileName, scopeName, kinds):
        sql = "select * from tags where file = '" + fileName + "' " \
                + " and scope='" + scopeName + "' "

        if kinds:
            sql += " and kind in("
            for i in kinds:
                sql += "'" + i + "',"
            sql = sql[:-1] + ")"

        return DoFetchTags(sql)

    def GetAllTagsNames(self):
        names = []
        try:
            sql = "SELECT distinct name FROM tags order by name ASC LIMIT " \
                    + str(self.GetMaxWorkspaceTagToColour())

            for row in self.Query(sql):
                # add unique strings only
                names.append(row[0])
        except:
            pass

        return names

    def GetTagsNames(self, kinds):
        if not kinds:
            return []

        names = []
        try:
            whereClause = " kind IN ("
            for kind in kinds:
                whereClause += "'" + kind + "',"

            whereClause = whereClause[:-1] + ") "

            sql = "SELECT distinct name FROM tags WHERE "
            sql += whereClause + " order by name ASC LIMIT " \
                    + str(self.GetMaxWorkspaceTagToColour())
            for row in self.Query(sql):
                names.append(row[0])
        except:
            pass

        return names

    def GetTagsByScopesAndKinds(self, scopes, kinds):
        if not kinds or not scopes:
            return []

        sql = "select * from tags where scope in ("
        for scope in scopes:
            sql += "'" + scope + "',"
        sql = sql[:-1] + ") "

        return self.DoFetchTags(sql, kinds)

    def GetGlobalFunctions(self):
        sql = "select * from tags where scope = '<global>' "\
                "AND kind IN ('function', 'prototype') LIMIT " \
                + str(self.GetSingleSearchLimit())
        return self.DoFetchTags(sql)

    def GetTagsByFiles(self, files):
        if not files:
            return []

        sql = "select * from tags where file in ("
        for file in files:
            sql += "'" + file + "',"
        sql = sql[:-1] + ")"
        return self.DoFetchTags(sql)

    def GetTagsByFilesAndScope(self, files, scope):
        if not files:
            return []

        sql = "select * from tags where file in ("
        for file in files:
            sql += "'" + file + "',"
        sql = sql[:-1] + ")"
        sql += " AND scope='" + scope + "'"
        return self.DoFetchTags(sql)

    def GetTagsByFilesKindsAndScope(self, files, kinds, scope):
        if not files:
            return []

        sql = "select * from tags where file in ("
        for file in files:
            sql += "'" + file + "',"
        sql = sql[:-1] + ")"

        sql += " AND scope='" + scope + "'"

        return self.DoFetchTags(sql, kinds)

    def GetTagsByFilesScopeTyperefAndKinds(self, files, kinds, scope, typeref):
        if not files:
            return []

        sql = "select * from tags where file in ("
        for file in files:
            sql += "'" + file + "',"
        sql = sql[:-1] + ")"

        sql += " AND scope='" + scope + "'"
        sql += " AND typeref='" + typeref + "'"

        return self.DoFetchTags(sql, kinds)

    def GetTagsByKindsLimit(self, kinds, orderingColumn, order, limit, partName):
        sql = "select * from tags where kind in ("
        for kind in kinds:
            sql += "'" + kind + "',"
        sql = sql[:-1] + ") "

        if orderingColumn:
            sql += "order by " + orderingColumn
            if order == ITagsStorage.OrderAsc:
                sql += " ASC"
            elif order == ITagsStorage.OrderDesc:
                sql += " DESC"
            else:
                pass

        if partName:
            tmpName = partName.replace('_', '^_')
            sql += " AND name like '%%" + tmpName + "%%' ESCAPE '^' "

        if limit > 0:
            sql += " LIMIT " + str(limit)

        return self.DoFetchTags(sql)

    def IsTypeAndScopeExistLimitOne(self, typeName, scope):
        path = ''

        # Build the path
        if scope and scope != "<global>":
            path += scope + "::"

        path += typeName
        sql += "select ID from tags where path='" + path \
                + "' and kind in ('class', 'struct', 'typedef') LIMIT 1"

        try:
            for row in self.Query(sql):
                return True
        except:
            pass

        return False

    def GetDereferenceOperator(self, scope):
        sql = "select * from tags where scope ='" + scope \
                + "' and name like 'operator%->%' LIMIT 1"
        return self.DoFetchTags(sql)

    def GetSubscriptOperator(self, scope):
        sql = "select * from tags where scope ='" + scope \
                + "' and name like 'operator%[%]%' LIMIT 1"
        return self.DoFetchTags(sql)

    def ClearCache(self):
        if self.cache:
            self.cache.Clear()

#===============================================================================
    # codelite 的 macro 处理, 暂不实现

    def GetMacro(self, name):
        # TODO: PPToken
        token = None

        try:
            sql = "select * from MACROS where name = '" + name + "'"
            for row in self.Query(sql):
                token = self.PPTokenFromSQlite3ResultSet(row)
                return token
        except:
            pass
        return token

    def StoreMacros(self, table):
        # TODO: PPToken
        pass

    def GetMacrosDefined(self, files, usedMacros):
        if not files or not usedMacros:
            return []

        defMacros = []

        # Create the file list SQL string, used for IN operator
        sFileList = ""
        for file in files:
            sFileList += "'" + file + "',"
        if sFileList:
            sFileList = sFileList[:-1]

        # Create the used macros list SQL string, used for IN operator
        sMacroList = ""
        for usedMacro in usedMacros:
            sMacroList += "'" + usedMacro + "',"
        if sMacroList:
            sMacroList = sMacroList[:-1]

        try:
            # Step 1 : Retrieve defined macros in MACROS table
            sql = "select name from MACROS where file in (" + sFileList + ")" \
                    + " and name in (" + sMacroList + ")"
            for row in self.db.execute(sql):
                defMacros.append(row[0])

            # Step 2 : Retrieve defined macros in SIMPLE_MACROS table
            sql = "select name from SIMPLE_MACROS where file in (" + sFileList \
                    + ")" + " and name in (" + sMacroList + ")"
            for row in self.db.execute(sql):
                defMacros.append(row[0])
        except:
            pass

        return defMacros




#CTAGS = 'ctags'
CTAGS = os.path.expanduser('~/bin/vlctags')
CTAGS_OPTS = '--excmd=pattern --sort=no --fields=aKmSsnit '\
        '--c-kinds=+p --c++-kinds=+p'

def ParseFiles(files):
    '返回标签文本'
    if not files:
        return ''

    tags = ''
    #for f in files:
        # TODO: 路径要不要转为绝对路径?
        #cmd = "%s %s -f - '%s'" % (CTAGS, CTAGS_OPTS, f)
        #tags += os.popen(cmd).read()
    cmd = "%s %s -f - '%s'" % (CTAGS, CTAGS_OPTS, "' '".join(files))
    tags = os.popen(cmd).read()
    return tags

def ParseFile(fileName):
    return ParseFiles([fileName])

def ParseFilesAndStore(files, storage, filterNonNeed = True):
    import time

    # 全部转为绝对路径
    tmpFiles = [os.path.abspath(f) for f in files]

    # 确保打开了一个数据库
    if not storage.OpenDatabase():
        return

    # 过滤不需要的. 通过比较时间戳
    if filterNonNeed:
        filesMap = storage.GetFilesMap()
        mapLen = len(tmpFiles)
        idx = 0
        while idx < mapLen:
            f = tmpFiles[idx]
            if filesMap.has_key(f):
                # 开始比较时间戳
                if int(os.path.getmtime(f)) \
                   < filesMap[f].GetLastRetaggedTimestamp():
                    # 过滤掉
                    del tmpFiles[idx]
                    mapLen -= 1
                    continue
            idx += 1

    tags = ParseFiles(tmpFiles)
    storage.Store(tags)
    for f in tmpFiles:
        if os.path.isfile(f):
            storage.InsertFileEntry(f, int(time.time()))
        # 文件路径全部转为绝对路径
        #absFile = os.path.abspath(f)
        #if os.path.isfile(absFile):
            #storage.InsertFileEntry(absFile, int(time.time()))




if __name__ == '__main__':
    #print 'enter'
    li = ['main.h', 'main.cpp']
    #tags = ParseFiles(li)
    #print '-' * 40
    #print tags
    #storage = TagsStorageSQLite()
    #storage.OpenDatabase('TestTags2.db')
    #storage.Store(tags)

    storage = TagsStorageSQLite()
    storage.OpenDatabase('TestTags3.db')
    ParseFilesAndStore(li, storage)
    #print storage.GetTagsByScopeAndKind('DFoo', 'member')
    entries = storage.GetTagsByScopesAndKinds(
        ['DFoo', 'Foo', 'Foo2'], ['member', 'function'])
    entries = storage.GetTagsByScopeAndName('DFoo', '', True)
    print len(entries)
    for entry in entries:
        entry.Print()


autoload/omnicpp/complete.vim	[[[1
377
" Description:  Omni completion script for resolve namespace
" Maintainer:   fanhe <fanhed@163.com>
" Last Change:  2011 May 14
" License:      GPLv2

"临时启用选项函数 {{{2
function! s:SetOpts()
    let s:bak_cot = &completeopt

    if g:VLOmniCpp_ItemSelectionMode == 0 " 不选择
        set completeopt-=menu,longest
        set completeopt+=menuone
    elseif g:VLOmniCpp_ItemSelectionMode == 1 " 选择并插入文本
        set completeopt-=menuone,longest
        set completeopt+=menu
    elseif g:VLOmniCpp_ItemSelectionMode == 2 " 选择但不插入文本
        set completeopt-=menu,longest
        set completeopt+=menuone
    else
        set completeopt-=menu
        set completeopt+=menuone,longest
    endif

    return ''
endfunction
function! s:RestoreOpts()
    if exists('s:bak_cot')
        let &completeopt = s:bak_cot
        unlet s:bak_cot
    else
        return ""
    endif

    let ret = ""

    if pumvisible()
        if g:VLOmniCpp_ItemSelectionMode == 0 " 不选择
            let ret = "\<C-p>"
        elseif g:VLOmniCpp_ItemSelectionMode == 1 " 选择并插入文本
            let ret = ""
        elseif g:VLOmniCpp_ItemSelectionMode == 2 " 选择但不插入文本
            let ret = "\<C-p>\<Down>"
        else
            let ret = "\<Down>"
        endif
    endif

    return ret
endfunction
function! s:CheckIfSetOpts()
    let l:line = getline('.')
    let l:col = col('.') - 1
    "若是成员补全，添加 longest
    if l:line[l:col-2:] =~ '->' || l:line[l:col-1:] =~ '\.' 
                \|| l:line[l:col-2:] =~ '::'
        call s:SetOpts()
    endif

    return ''
endfunction
"}}}
function! s:ShouldComplete() "{{{2
    if (getline('.') =~ '#\s*\(include\|import\)')
        "写头文件，忽略
        return 0
    else
        "检测光标所在的区块，如果在注释、双引号、浮点数？时，忽略
        let nLine = line('.')
        let nCol = col('.') - 1
        if nCol < 1
            return 0
        endif
        for l:id in synstack(nLine, nCol)
            if match(synIDattr(l:id, 'name'), '\CComment\|String\|Number')
                        \ != -1
                return 0
            endif
        endfor
        return 1
    endif
endfunction
"}}}
function! s:LaunchOmnicppCompletion() "{{{2
    if s:ShouldComplete()
        return "\<C-x>\<C-o>"
    else
        return ''
    endif
endfunction
"}}}
function! s:CompleteByChar(char) "{{{2
    if a:char ==# '.'
        return a:char . s:LaunchOmnicppCompletion()
    elseif a:char ==# '>'
        if getline('.')[col('.') - 2] != '-'
            return a:char
        else
            return a:char . s:LaunchOmnicppCompletion()
        endif
    elseif a:char ==# ':'
        if getline('.')[col('.') - 2] != ':'
            return a:char
        else
            return a:char . s:LaunchOmnicppCompletion()
        endif
    endif
endfunction
"}}}
function! omnicpp#complete#Init() "{{{2
    " 初始化数据库
    call VimTagsManagerInit()

    " 初始化设置
    call omnicpp#settings#Init()

    "setlocal completefunc=
    setlocal omnifunc=omnicpp#complete#Complete

    if g:VLOmniCpp_MayCompleteDot
        inoremap <silent> <buffer> . 
                    \<C-r>=<SID>SetOpts()<CR>
                    \<C-r>=<SID>CompleteByChar('.')<CR>
                    \<C-r>=<SID>RestoreOpts()<CR>
    endif

    if g:VLOmniCpp_MayCompleteArrow
        inoremap <silent> <buffer> > 
                    \<C-r>=<SID>SetOpts()<CR>
                    \<C-r>=<SID>CompleteByChar('>')<CR>
                    \<C-r>=<SID>RestoreOpts()<CR>
    endif

    if g:VLOmniCpp_MayCompleteColon
        inoremap <silent> <buffer> : 
                    \<C-r>=<SID>SetOpts()<CR>
                    \<C-r>=<SID>CompleteByChar(':')<CR>
                    \<C-r>=<SID>RestoreOpts()<CR>
    endif

    inoremap <script> <Plug>SmartComplete 
                \<C-r>=<SID>CheckIfSetOpts()<CR>
                \<C-r>=<SID>LaunchOmnicppCompletion()<CR>
                \<C-r>=<SID>RestoreOpts()<CR>

    if g:VLOmniCpp_ItemSelectionMode > 4
        imap <silent> <buffer> <C-n> <Plug>SmartComplete
    else
        "inoremap <silent> <buffer> <C-n> 
                    "\<C-r>=<SID>SetOpts()<CR>
                    "\<C-r>=<SID>LaunchOmnicppCompletion()<CR>
                    "\<C-r>=<SID>RestoreOpts()<CR>
    endif
endfunction
"}}}
let s:CompletionType_NormalCompl = 0    " 作用域内符号补全
let s:CompletionType_MemberCompl = 1    " 作用域内变量成员补全 ('::', '->', '.')
let s:bDoNotComplete = 0 " 用于 omnifunc 第一阶段与第二阶段通讯, 指示禁止补全
" omnifunc
function! omnicpp#complete#Complete(findstart, base) "{{{2
    if a:findstart
        "call g:Timer.Start() "计时用
        let szLine = getline('.')
        let startIdx = col('.') - 1

        " 跳过光标在注释和字符串中的补全请求
        if omnicpp#utils#IsCursorInCommentOrString()
            " BUG: 返回 -1 仍然进入下一阶段
            let s:bDoNotComplete = 1
            return -1
        endif

        let b:completionType = s:CompletionType_NormalCompl

        " TODO: 应该分析 tokens
        if szLine[startIdx - 1] =~ '\s' || szLine == ''
            " 光标前的字符为空字符, 要么不能补全, 要么是无预输入的成员补全
            let b:completionType = s:CompletionType_MemberCompl

            "跳过空格， . -> :: 之间是允许空格的...
            while startIdx > 0 && szLine[startIdx - 1] =~ '\s'
                let startIdx -= 1
            endwhile

            if szLine[:startIdx] =~# '->$\|::$\|\.$'
                " 成员补全. eg. std::, MyClass->
                let b:col = startIdx + 1
                return col('.')
            else
                " 没有成员补全操作符, 补全失败
                let s:bDoNotComplete = 1
                return -1
            endif
        else
            " 光标前的字符为非空字符, 无论如何都会请求补全
            " 两种补全都有可能. eg. MyClass->|, MyClass.m_|
            if szLine[: startIdx-1] =~# '->$\|::$\|\.$'
                " 无预输入的成员补全. eg. std::, MyClass->
                let b:col = startIdx + 1
                let b:completionType = s:CompletionType_MemberCompl
                return col('.')
            else
                " 有预输入的作用域符号或成员补全
                if match(szLine, '->\|::\|\.') >= 0
                    let b:completionType = s:CompletionType_MemberCompl
                endif

                return match(szLine, '\<\w\+$')
            endif
        endif
    endif

    let result = []

    if s:bDoNotComplete
        let s:bDoNotComplete = 0
        return result
    endif

    let base = a:base
    let b:base = a:base
    let b:findstart = col('.')

    let scopeStack = omnicpp#scopes#GetScopeStack()
    let simpleSS = omnicpp#complete#SimplifyScopeStack(scopeStack)
    let tags = []
    let b:tags = tags " 调试用

    if b:completionType == s:CompletionType_NormalCompl
        " (1) 作用域内符号补全模式

        " 与 TagStorage 一致的 scope
        let tagScope = omnicpp#complete#GetTagScopeFromScopeStack(scopeStack, 1)
        let tagScopes = [tagScope]

        " 添加全局 scope
        if tagScope != '<global>'
            let tagScopes += ['<global>']
        endif

        let tags = g:GetTagsByScopesAndName(tagScopes, base)
        "let result = s:BuildPopupMenuFromTags(tags)
    else
        " (2) 作用域内成员补全模式

        let omniSS = omnicpp#resolvers#GetOmniScopeStack()
        let tagScopes = omnicpp#resolvers#ResolveOmniScopeStack(
                    \scopeStack, omniSS)
        let b:omniSS = omniSS
        let b:tagScopes = tagScopes

        let tags = g:GetTagsByScopesAndName(tagScopes, base)
    endif


    call map(tags, 's:ExtendTagToPopupMenuItem(v:val)')
    let result = tags

    "call g:Timer.EndEchoMes() " 计时用

    return result
endfunc
"}}}
" 简化 scopeStack, scope.kind 为 'other' 的合并进前一个 scope
" 返回的是副本
function! omnicpp#complete#SimplifyScopeStack(scopeStack) "{{{2
    let result = []
    for scope in a:scopeStack
        if scope.kind == 'other'
            let prevScope = result[-1]
            let prevScope.namespaces += scope.namespaces
            let prevScope.includes += scope.namespaces
        else
            call add(result, copy(scope))
        endif
    endfor

    return result
endfunc
"}}}
function! s:BuildPopupMenuFromTags(tags) "{{{2
    let tags = a:tags
endfunc
"}}}
" Extend a tag entry to a popup entry
" 从 tag 字典生成补全字典(用于补全条目), 并把结果存在 tag 字典中(key 不重复?)
function! s:ExtendTagToPopupMenuItem(tag) "{{{2
    let tag = a:tag

    " Add the access
    let szMenu = ''
    let accessChar = {'public': '+','protected': '#','private': '-'}
    if 1
        if has_key(tag, 'access') && has_key(accessChar, tag.access)
            let szMenu = szMenu.accessChar[tag.access]
        else
            let szMenu = szMenu." "
        endif
    endif

    " Formating optional menu string we extract the scope information
    let szName = tag.name
    let szWord = szName
    let szAbbr = szName

    if 1
        let szMenu = szMenu . ' ' . tag.parent
    endif

    " Formating information for the preview window
    if index(['f', 'p'], tag.kind[0]) >= 0
        if 0 && has_key(tag, 'signature')
            let szAbbr .= tag.signature
        else
            let szAbbr .= '()'
        endif
    endif

    let szInfo = ''

    " If a function is a ctor we add a new key in the tag
    " 构造函数处理, 以及友元处理
    "if index(['f', 'p'], tag.kind[0]) >= 0
        "if match(szName, '^\~') < 0 && a:szTypeName =~ '\C\<'.szName.'$'
            " It's a ctor
            "let tag['ctor'] = 1
        "elseif has_key(tag, 'access') && tag.access == 'friend'
            " Friend function
            "let tag['friendfunc'] = 1
        "endif
    "endif

    " Extending the tag item to a popup item
    let tag['word']     = szWord
    let tag['abbr']     = szAbbr
    let tag['menu']     = szMenu
    let tag['info']     = szInfo
    let tag['icase']    = &ignorecase
    let tag['dup']      = 0
    "let tag['dup'] = (s:hasPreviewWindow 
                "\&& index(['f', 'p', 'm'], tag.kind[0]) >= 0)
    return tag
endfunc
"}}}
" 如果传入第二个参数且非 0, 剔除最后的函数作用域
function omnicpp#complete#GetTagScopeFromScopeStack(scopeStack, ...) "{{{2
    let result = []
    let scopeStackLen = len(a:scopeStack)

    let idx = 0
    while idx < scopeStackLen
        let scope = a:scopeStack[idx]
        if scope.kind == 'other'
            break
        elseif scope.kind == 'file'
            " 跳过
        else
            call add(result, scope.name)
        endif

        let idx += 1
    endwhile

    " TagStorage 不支持函数局部变量, 所以如果最后的作用域是函数, 剔除之...
    if a:0 > 0 && a:1 != 0 && !empty(result)
        if a:scopeStack[idx-1].kind == 'function'
            call remove(result, -1)
        endif
    endif

    if empty(result)
        call add(result, '<global>')
    endif

    return join(result, '::')
endfunc
"}}}
" vim:fdm=marker:fen:expandtab:smarttab:fdl=1:
autoload/omnicpp/utils.vim	[[[1
500
" Description:	Omni completion utils script
" Maintainer:   fanhe <fanhed@163.com>
" Last Change:  2011 May 11
" License:      GPLv2


" 过滤所有有 class, struct, union, namespace 键值的项
" 如果是枚举, 只过滤枚举值不匹配 '^\w\+$' 的项
" Note: 用数据库是应该不需要
let g:omnicpp#utils#szFilterGlobalScope = "(!has_key(v:val, 'class') "
            \."&& !has_key(v:val, 'struct') "
            \."&& !has_key(v:val, 'union') "
            \."&& !has_key(v:val, 'namespace')"
let g:omnicpp#utils#szFilterGlobalScope .= " && (!has_key(v:val, 'enum') "
            \."|| (has_key(v:val, 'enum') && v:val.enum =~ '^\\w\\+$')))"

" Expression used to ignore comments
" Note: this expression drop drastically the performance
" 检查语法项目真的这么慢吗?
"let omnicpp#utils#expIgnoreComments = 
            "\'match(synIDattr(synID(line("."), col("."), 1), "name"), '
            "\'"\CcComment")!=-1'
" This one is faster but not really good for C comments
let omnicpp#utils#reIgnoreComment = escape('\/\/\|\/\*\|\*\/', '*/\')
let omnicpp#utils#expIgnoreComments = 
            \'getline(".") =~ g:omnicpp#utils#reIgnoreComment'

" Characters to escape in a filename for vimgrep
"TODO: Find more characters to escape
let omnicpp#utils#szEscapedCharacters = ' %#'

" 比较两个光标位置 
function! omnicpp#utils#CmpPos(pos1, pos2) "{{{2
	let pos1 = a:pos1
	let pos2 = a:pos2
	if pos1[0] > pos2[0]
		return 1
	elseif pos1[0] < pos2[0]
		return -1
	else
		if pos1[1] > pos2[1]
			return 1
		elseif pos1[1] < pos2[1]
			return -1
		else
			return 0
		endif
	endif
endfunc
"}}}
" 获取当前语句的开始位置, 若当前语句为空, 返回 [0, 0]
" Note: 
"   1. 当前语句的开始位置可能是预处理的行或者字符串或者注释
"      若想剔除预处理的行, 可在获取代码的时候过滤掉
"   2. 返回的位置可能在原始位置的后面, 需要调用者检查
function! omnicpp#utils#GetCurStatementStartPos(...) "{{{2
    let origCursor = getpos('.')
    let origPos = origCursor[1:2]

    let result = [0, 0]

    while searchpos('[;{}]\|\%^', 'bW') != [0, 0]
        if omnicpp#utils#IsCursorInCommentOrString()
            continue
        else
            break
        endif
    endwhile

    if getpos('.')[1:2] == [1, 1] " 到达文件头的话, 接受光标所在位置的匹配
        let result = searchpos('[^ \t]', 'Wc')
    else
        let result = searchpos('[^ \t]', 'W')
    endif

    if a:0 > 0 && a:1
        " 如果传入参数且非零, 移动光标
    else
        call setpos('.', origCursor)
    endif

    return result
endfunc
"}}}
" Get code without comments and with empty strings
" szSingleLine must not have carriage return
" 获取行字符串中的有效 C/C++ 代码, 双引号中的字符串会被清空
" 要求字符串中不能有换行符, 亦即只能处理单行的字符串
" TODO: 清理 # 预处理
function! omnicpp#utils#GetCodeFromLine(szSingleLine) "{{{2

    " 清理预处理
    if match(a:szSingleLine, '^\s*#') >= 0
        return ''
    endif

    " We set all strings to empty strings, it's safer for 
    " the next of the process
	" TODO: '"\"", foo, "foo"'
	" 方法是替换不匹配 \" 的 " 的最小长度的匹配
    let szResult = substitute(a:szSingleLine, '".*"', '""', 'g')

    " Removing C++ comments, we can use the pattern ".*" because
    " we are modifying a line
    let szResult = substitute(szResult, '\/\/.*', '', 'g')

    " Now we have the entire code in one line and we can remove C comments
    return s:RemoveCComments(szResult)
endfunc
"}}}
" Remove C comments on a line
function! s:RemoveCComments(szLine) "{{{2
    let result = a:szLine

    " We have to match the first '/*' and first '*/'
    let startCmt = match(result, '\/\*')
    let endCmt = match(result, '\*\/')
    while startCmt != -1 && endCmt != -1 && startCmt < endCmt
        if startCmt > 0
            let result = result[ : startCmt-1 ] . result[ endCmt+2 : ]
        else
            " Case where '/*' is at the start of the line
            let result = result[ endCmt+2 : ]
        endif
        let startCmt = match(result, '\/\*')
        let endCmt = match(result, '\*\/')
    endwhile
    return result
endfunc
"}}}
" Get a c++ code from current buffer from [lineStart, colStart] to 
" [lineEnd, colEnd] without c++ and c comments, without end of line
" and with empty strings if any
" @return a string
function! omnicpp#utils#GetCode(posStart, posEnd) "{{{2
	"TODO: 处理反斜杠续行
    let posStart = a:posStart
    let posEnd = a:posEnd
    if a:posStart[0] > a:posEnd[0]
        let posStart = a:posEnd
        let posEnd = a:posStart
    elseif a:posStart[0] == a:posEnd[0] && a:posStart[1] > a:posEnd[1]
        let posStart = a:posEnd
        let posEnd = a:posStart
    endif

    " Getting the lines
    let lines = getline(posStart[0], posEnd[0])
    let lenLines = len(lines)

    " Formatting the result
    let result = ''
    if lenLines == 1
        let sStart = posStart[1] - 1
        let sEnd = posEnd[1] - 1
        let line = lines[0]
        let lenLastLine = strlen(line)
        let sEnd = (sEnd > lenLastLine) ? lenLastLine : sEnd
        if sStart >= 0
            let result = omnicpp#utils#GetCodeFromLine(line[ sStart : sEnd ])
        endif
    elseif lenLines > 1
        let sStart = posStart[1] - 1
        let sEnd = posEnd[1] - 1
        let lenLastLine = strlen(lines[-1])
        let sEnd = (sEnd > lenLastLine)?lenLastLine : sEnd
        if sStart >= 0
            let lines[0] = lines[0][ sStart : ]
            let lines[-1] = lines[-1][ : sEnd ]
            for aLine in lines
                let result = result . omnicpp#utils#GetCodeFromLine(aLine)." "
            endfor
            let result = result[:-2]
        endif
    endif

    " Now we have the entire code in one line and we can remove C comments
    return s:RemoveCComments(result)
endfunc
"}}}
" Check if the cursor is in comment or string
" 根据语法检测光标是否在注释或字符串中或 include 预处理中
" FIXME: 真的需要检查预处理?!
function! omnicpp#utils#IsCursorInCommentOrString() "{{{2
    " FIXME: case. |"abc" . 判定为在字符串中
    "return match(synIDattr(synID(line("."), col(".")-1, 1), "name"), 
				"\'\C\<cComment\|\<cCppString\|\<cIncluded') >= 0
    for id in synstack(line('.'), col('.'))
        if match(synIDattr(id, 'name'), 
                    \'\CComment\|String\|Included') != -1
            return 1
        endif
    endfor
    return 0
endfunc
"}}}
" Tokenize the current instruction until the cursor position.
" Param:    可选参数为附加字符串, 或控制是否简化代码为 omni 分析用 tokens
" Return:   list of tokens
function! omnicpp#utils#TokenizeStatementBeforeCursor(...) "{{{2
    let szAppendText = ''
    let bSimplify = 0
    if a:0 > 0 && type(a:1) == type('')
        let szAppendText = a:1
    elseif a:0 > 0 && type(a:1) == type(0)
        let bSimplify = a:1
    endif

    " TODO: 排除注释和字符串中的匹配
    let startPos = searchpos('[;{}]\|\%^', 'bWn')
    let curPos = getpos('.')[1:2]
    " We don't want the character under the cursor
    let column = curPos[1]-1
    let curPos[1] = (column < 1) ? 1 : column
    " Note: [1:] 剔除了上一条语句的结束符
    let szCode = omnicpp#utils#GetCode(startPos, curPos)[1:] . szAppendText
    if bSimplify
        let szCode = omnicpp#utils#SimplifyCodeForOmni(szCode)
    endif
    return omnicpp#tokenizer#Tokenize(szCode)
endfunc
"}}}
function! omnicpp#utils#GetCurStatementBeforeCursor(...) "{{{
    let bSimplify = 0
    if a:0 > 0
        let bSimplify = a:1
    endif

    let curPos = getpos('.')[1:2]
    let startPos = omnicpp#utils#GetCurStatementStartPos()
    " 不包括光标下的字符
    let szCode = omnicpp#utils#GetCode(startPos, curPos)[:-2]
    if bSimplify
        let szCode = omnicpp#utils#SimplifyCodeForOmni(szCode)
    endif

    return szCode
endfunc
"}}}
" 简化 omni 补全请求代码, 效率不是太高
" eg1. A::B("\"")->C(" a(\")z ", ')'). -> A::B()->C().
" eg2. ((A*)(B))->C. -> ((A*)B)->C.
" egx. A::B(C.D(), ((E*)F->G)).
function! omnicpp#utils#SimplifyCodeForOmni(szOmniCode) "{{{2
	let s = a:szOmniCode
	
	" 1. 清理函数函数
	let s = s:StripFuncArgs(s)
	" 2. 清理多余的括号, 不会检查匹配的情况. eg. ((A*)((B))) -> ((A*)B)
	" TODO: 需要更准确
	let s = substitute(s, '\W\zs(\+\(\w\+\))\+\ze', '\1', 'g')

	return s
endfunc
"}}}
" 剔除所有函数参数, 仅保留一个括号
function! s:StripFuncArgs(szOmniCode) "{{{2
	let s = a:szOmniCode

	" 1. 清理引号内的 \" , 因为正则貌似是不能直接匹配 ("\"")
	let s = substitute(s, '\\"', '', 'g')
	" 2. 清理双引号的内容, 否则双引号内的括号会干扰括号替换
	let s = substitute(s, '".\{-}"', '', 'g')
	" 3. 清理 '(' 和 ')', 避免干扰
	let s = substitute(s, "'('\\|')'", '', 'g')

	let szResult = ''
	let nStart = 0
	while nStart < len(s) && nStart != -1
		let nEnd = matchend(s, '\w\s*(', nStart)
		if nEnd != -1
			let szResult .= s[nStart : nEnd - 1]

			let nStart = nEnd
			let nCount = 1
			" 开始寻找匹配的 ) 的位置
			for i in range(nStart, len(s) - 1)

				let c = s[i]

				if c == '('
					let nCount += 1
				elseif c == ')'
					let nCount -= 1
				endif

				if nCount == 0
					let nStart = i + 1
					let szResult .= ')'
					break
				endif
			endfor

		else
			let szResult .= s[nStart :]
			break
		endif
	endwhile

	return szResult
endfunc
"}}}
" Tokenize the current instruction until the word under the cursor.
" @return list of tokens
function! omnicpp#utils#TokenizeCurrentInstructionUntilWord()
    let startPos = searchpos('[;{}]\|\%^', 'bWn')

    " Saving the current cursor pos
    let originalPos = getpos('.')

    " We go at the end of the word
    execute 'normal gee'
    let curPos = getpos('.')[1:2]

    " Restoring the original cursor pos
    call setpos('.', originalPos)

    let szCode = omnicpp#utils#GetCode(startPos, curPos)[1:]
    return omni#cpp#tokenizer#Tokenize(szCode)
endfunc

" Build parenthesis groups
" add a new key 'group' in the token
" where value is the group number of the parenthesis
" eg: (void*)(MyClass*)
"      group1  group0
" if a parenthesis is unresolved the group id is -1      
" Return: a copy of a:tokens with parenthesis group
function! omnicpp#utils#BuildParenthesisGroups(tokens)
    let tokens = copy(a:tokens)
    let kinds = {'(': '()', ')' : '()', '[' : '[]', ']' : '[]', '<' : '<>', 
				\'>' : '<>', '{': '{}', '}': '{}'}
    let unresolved = {'()' : [], '[]': [], '<>' : [], '{}' : []}
    let groupId = 0

    " Note: we build paren group in a backward way
    " because we can often have parenthesis unbalanced statement
    " eg: doSomething(_member.get()->
    for token in reverse(tokens)
        if index([')', ']', '>', '}'], token.value) >= 0
            let token['group'] = groupId
            call extend(unresolved[kinds[token.value]], [token])
            let groupId += 1
        elseif index(['(', '[', '<', '{'], token.value)>=0
            if len(unresolved[kinds[token.value]])
                let tokenResolved = remove(unresolved[kinds[token.value]], -1)
                let token['group'] = tokenResolved.group
            else
                let token['group'] = -1
            endif
        endif
    endfor

    return reverse(tokens)
endfunc

" Determine if tokens represent a C cast
" @return
"   - itemCast
"   - itemCppCast
"   - itemVariable
"   - itemThis
function! omnicpp#utils#GetCastType(tokens)
    " Note: a:tokens is not modified
    let tokens = omnicpp#utils#SimplifyParenthesis(
				\omnicpp#utils#BuildParenthesisGroups(a:tokens))

    if tokens[0].value == '('
        return 'itemCast' 
    elseif index(['static_cast', 'dynamic_cast', 'reinterpret_cast', 
                \'const_cast'], tokens[0].value)>=0
        return 'itemCppCast'
    else
        for token in tokens
            if token.value=='this'
                return 'itemThis'
            endif
        endfor
        return 'itemVariable' 
    endif
endfunc

" Remove useless parenthesis
" 剔除不必要的括号. 主要用于解析强制类型转换的 tokens
function! omnicpp#utils#SimplifyParenthesis(tokens)
    "Note: a:tokens is not modified
    let tokens = a:tokens
    " We remove useless parenthesis eg: (((MyClass)))
    if len(tokens)>2
        while tokens[0].value == '(' 
					\&& tokens[-1].value == ')' 
					\&& tokens[0].group == tokens[-1].group
            let tokens = tokens[1:-2]
        endwhile
    endif
    return tokens
endfunc


" Extract type from tokens.
" eg: examples of tokens format
"   'const MyClass&'
"   'const map < int, int >&'
"   'MyNs::MyClass'
"   '::MyClass**'
"   'MyClass a, *b = NULL, c[1] = {};
"   'hello(MyClass a, MyClass* b'
" @return the type info string eg: ::std::map
" can be empty
function! omnicpp#utils#GetVariableType(tokens)
    if type(a:tokens) == type('')
        let tokens = omnicpp#tokenizer#Tokenize(a:tokens)
    else
        let tokens = a:tokens
    endif

    let szResult = ''
    let state = 0

    let tokens = omnicpp#utils#BuildParenthesisGroups(tokens)

    " If there is an unbalanced parenthesis we are in a parameter list
    let bParameterList = 0
    for token in tokens
        if token.value == '(' && token.group==-1
            let bParameterList = 1
            break
        endif
    endfor

    if bParameterList
        let tokens = reverse(tokens)
        let state = 0
        let parenGroup = -1
        for token in tokens
            if state == 0
                if token.value == '>'
                    let parenGroup = token.group
                    let state = 1
                elseif token.kind == 'cppWord'
                    let szResult = token.value.szResult
                    let state = 2
                elseif index(['*', '&'], token.value)<0
                    break
                endif
            elseif state == 1
                if token.value == '<' && token.group == parenGroup
                    let state=0
                endif
            elseif state == 2
                if token.value == '::'
                    let szResult = token.value.szResult
                    let state=3
                else
                    break
                endif
            elseif state == 3
                if token.kind == 'cppWord'
                    let szResult = token.value.szResult
                    let state = 2
                else
                    break
                endif
            endif
        endfor
        return szResult
    endif

    for token in tokens
        if state == 0
            if token.value == '::'
                let szResult .= token.value
                let state = 1
            elseif token.kind == 'cppWord'
                let szResult .= token.value
                let state = 2
                " Maybe end of token
            endif
        elseif state == 1
            if token.kind == 'cppWord'
                let szResult .= token.value
                let state = 2
                " Maybe end of token
            else
                break
            endif
        elseif state == 2
            if token.value == '::'
                let szResult .= token.value
                let state = 1
            else
                break
            endif
        endif
    endfor
    return szResult
endfunc


" vim:fdm=marker:fen:expandtab:smarttab:fdl=1:
autoload/omnicpp/scopes.vim	[[[1
503
" Description:  Omni completion script for resolve namespace
" Maintainer:   fanhe <fanhed@163.com>
" Last Change:  2011 May 13
" License:      GPLv2

" Note:  只解析当前文件的全局名空间和当前局部作用域的名空间, 不解析包含文件的
" 因为, 不应该在包含的文件使用名空间

" Note2: 只能解析具备良好风格习惯的代码, 因为始终不是编译器, 
" 即使有能力解析, 但是代价也会变大, 速度会下降

" scope 数据结构
" {
" 'kind':       <'file'|'container'|'function'|'other'>, 
" 'name':       <scope name>, 
" 'namespaces': [<ns1>, <ns2>, ...], 
" 'includes':   [<header1>, <header2>, ...]
" }
"
" 'kind' 的 'container' 意即 'namespace'|'class'|'struct'|'union'
" 'kind' 为 'file' 时, 'name' 为当前文件名(绝对路径).
" 'includes' 的列表项均为绝对路径. 一般只有在 'kind' 为 'file' 时才可能为非空.

" scopeStack 数据结构
" [<scope1>, <scope2>, ...]
"
" 首元素 scope 的 'kind' 必为 'file', 亦即, 首元素必为全局作用域
" 若视为栈的话, 底部为首元素, 顶部为尾元素

" 新建 scope 数据结构
function! s:NewScope() "{{{2
    return {'kind': '', 'name': '', 'namespaces': [], 'includes': []}
endfunc
"}}}

" 返回当前作用域栈
function! omnicpp#scopes#GetScopeStack() "{{{2
    " We store the cursor position because searchpairpos() moves the cursor
    let origCursor = getpos('.')
    let endPos = origCursor[1:2]
    let listCode = []
    let result = {'namespaces': [], 'scope': ''}
    let scopeStack = []

    " 名空间搜索的开始位置
    let searchStartPos = endPos[:]

	" 往上分析每个 {} 块, 以生成作用域栈
    while endPos != [0, 0]
        " {} 块的 { 位置
        let endPos = searchpairpos('{', '', '}', 'bW', 
                    \g:omnicpp#utils#expIgnoreComments)


        " 单独处理 for ( A; B; C ) 的情形
        let tmpCursor = getpos('.')
        let tmpPos = searchpos(')\s*$\|)\s*{\s*$', 'Wb')
        if tmpPos != [0, 0] && tmpCursor[1] - tmpPos[0] <= 1 " 必须相邻
            normal! %
            if getline('.')[:col('.')-1] =~# 'for\s*($'
                " 找到了 for
                let curScope = s:NewScope()
                let curScope.kind = 'other'
                let curScope.name = 'for'
                let scopeStack = [curScope] + scopeStack

                let searchStartPos = endPos[:]
                call setpos('.', tmpCursor)
                continue
            endif
        endif
        call setpos('.', tmpCursor)


        " FIXME: 无法获取 for 作用域. eg. for( A; B; C ) | { }
        let szReStartPos = '[;{}]\|\%^'
		"搜索 {} 块前的语句的开始位置, 用于获取之间的文本以分析
        let startPos = searchpairpos(szReStartPos, '', '{', 'bWn', 
                    \g:omnicpp#utils#expIgnoreComments)

        " If the file starts with a comment so the startPos can be [0,0]
        " we change it to [1,1]
        " 如果搜索失败, 定位光标至开始处
        " FIXME: 以下情况, 获取了两个包含文件的行
        " #include <stdio.h>
        " #include <stdlib.h>
        " class MyClass |{
        if startPos == [0, 0]
            let startPos = [1, 1]
        endif

        " Get lines backward from cursor position to last ; or { or }
        " or when we are at the beginning of the file.
        if endPos != [0, 0]
            " We remove the last character which is a '{'
            " We also remove starting { or } or ; if exits
			" 获取 {} 块前的一条有效语句代码并剔除前面和后面的多余字符
			" 结果 eg. '   int main(int argc, char **argv)'
            let szCodeWithoutComments = substitute(
                        \omni#cpp#utils#GetCode(startPos, endPos)[:-2], 
                        \'^[;{}]', '', 'g')
            let code = {'startLine' : startPos[0], 
                        \'code' : szCodeWithoutComments}

            let tokens = omnicpp#tokenizer#Tokenize(code.code)
            let curScope = s:NewScope()

            let tokensLen = len(tokens)
            let scopeType = 0 " 0 表示无名块类别
            let idx = 0
            while idx < tokensLen
                let token = tokens[idx]

                " 首先查找能确定类别的关键字符
                " 1. 容器类别 'namespace|class|struct|union'. eg. class A
                " 2. 作用域类别 '::'. eg. A B::C::D(E)
                " 3. 函数类别 '('. eg. A B(C)
                if token.kind == 'cppKeyword' 
                            \&& index(['namespace', 'class', 'struct', 'union'],
                            \   token.value) >= 0
                    " 容器类别
                    let scopeType = 1
                    let curScope.kind = 'container'
                    let curScope.name = tokens[idx+1].value
                    " 暂不支持在容器类型里使用名空间
                    "let curScope.namespaces = 
                    let scopeStack = [curScope] + scopeStack

                    break
                elseif token.kind == 'cppOperatorPunctuator' 
                            \&& token.value == '::'
                    " 作用域类别
                    let scopeType = 2
                    let tmpScopeStack = []

                    " FIXME: 不能处理释构函数 eg. A::~A()

                    " 无法分辨类型到底是 namespace 还是 class
                    let curScope.kind = 'container'
                    let curScope.name = tokens[idx-1].value
                    let tmpScopeStack = tmpScopeStack + [curScope]

                    " 继续分析
                    " 方法都是遇到操作符('::', '(')后确定前一个 token 的类别
                    let j = idx + 1
                    while j < tokensLen
                        let token = tokens[j]
                        if token.kind == 'cppOperatorPunctuator' 
                                    \&& token.value == '('

                            let curScope = s:NewScope()
                            if tokens[j-1].kind == 'cppKeyword'
                                " eg. if (), while (), switch () ...
                                let curScope.kind = 'other'
                            else
                                let curScope.kind = 'function'
                            endif
                            let curScope.kind = 'function'
                            let curScope.name = tokens[j-1].value
                            let curScope.namespaces 
                                        \= s:GetLocalUsingNamespacesByPos(
                                        \   searchStartPos)
                            " 添加方法与外部的循环是不同的
                            let tmpScopeStack =  tmpScopeStack + [curScope]

                            " 到了函数参数或条件判断位置, 已经完成
                            break
                        endif

                        if token.kind == 'cppOperatorPunctuator' 
                                    \&& token.value == '::'
                            let curScope = s:NewScope()
                            let curScope.kind = 'container'
                            let curScope.name = tokens[j-1].value
                            " 添加方法与外部的循环是不同的
                            let tmpScopeStack =  tmpScopeStack + [curScope]
                        endif

                        let j += 1
                    endwhile

                    let scopeStack = tmpScopeStack + scopeStack

                    break
                elseif token.kind == 'cppOperatorPunctuator' 
                            \&& token.value == '('
                    " 函数或条件类别
                    let scopeType = 3
                    if tokens[idx-1].kind == 'cppKeyword'
                        let curScope.kind = 'other'
                    else
                        let curScope.kind = 'function'
                    endif
                    let curScope.name = tokens[idx-1].value
                    " Note: namespaces 属性暂时只支持在函数中和文件中存在
                    let curScope.namespaces = s:GetLocalUsingNamespacesByPos(
                                \searchStartPos)
                    let scopeStack = [curScope] + scopeStack

                    break
                else
                    " TODO
                endif

                let idx += 1
            endwhile
        endif

        let searchStartPos = endPos[:]
    endwhile

    " Setting the cursor to the original position
    call setpos('.', origCursor)

    let globalScope = s:NewScope()
    let globalScope.kind = 'file'
    let globalScope.name = expand('%:p')
    " TODO: stopLine 的最小值为上述解析中的最后停留位置
    let globalScope.namespaces = s:GetGlobalUsingNamespaces(origCursor[1])
    "let globalScope.includes = omnicpp#includes#GetIncludeFiles()
    let scopeStack = [globalScope] + scopeStack

    return scopeStack
endfunc
"}}}

" Get namespaces used at the cursor postion in a vim buffer
" Note: The result depends on the current cursor position
" @return
"   -   List of namespace
function! omnicpp#scopes#GetUsingNamespaces() "{{{2
    " We have to get local using namespace declarations
    " We need the current cursor position and the position of the start of the
    " current scope

    " We store the cursor position because searchpairpos() moves the cursor
    let result = []

    " 1) Firstly, we have to get local using namespace declarations
    " 在局部作用域中搜索. 反向搜索
    let originalPos = getpos('.')
    let origPos = originalPos[1:2]

    let stopPos = s:GetStopPositionForLocalSearch()

    let stopLine = stopPos[0]
    let curPos = origPos
    let lastLine = 0 
    let nextStopLine = origPos[0]
    while curPos != [0, 0]
		" 搜索局部与某作用域(非作用于整个源文件)的 'using namespace' 声明
        let curPos = searchpos('\C}\|\(using\s\+namespace\)', 'bW', stopLine)
        if curPos != [0, 0] && curPos[0] != lastLine
            let lastLine = curPos[0]

            let szLine = getline('.')
            if origPos[0] == curPos[0]
                " We get the line until cursor position
                let szLine = szLine[:origPos[1]]
            endif

            let szLine = omni#cpp#utils#GetCodeFromLine(szLine)
            if match(szLine, '\Cusing\s\+namespace') < 0
                " We found a '}'
                let curPos = searchpairpos('{', '', '}', 'bW', 
                            \g:omnicpp#utils#expIgnoreComments)
            else
                " We get the namespace list from the line
                let result = s:GetNamespaceListFromLine(szLine) + result
                let nextStopLine = curPos[0]
            endif
        endif
    endwhile

    " Setting the cursor to the original position
    call setpos('.', originalPos)

    " 2) Now we can get all global using namespace declaration from the
    " 搜索全局的
    " beginning of the file to nextStopLine
    " nextStopLine 为光标原始位置, 或者刚刚搜索完局部声明的那个位置
    let result = s:GetGlobalUsingNamespaces(nextStopLine) + result

    " Resolving alias in the namespace list
    " TODO: For the next release
    "let g:omnicpp#scopes#CacheAlias= s:GetNamespaceAliasMap()
    "call s:ResolveAliasInNamespaceList(g:omnicpp#scopes#CacheAlias, result)

    return result
endfunc
"}}}

function! s:GetLocalUsingNamespacesByPos(pos) "{{{2
    let origCursor = getpos('.')

    call setpos('.', [0, a:pos[0], a:pos[1], 0])
    let result = s:GetLocalUsingNamespaces()

    call setpos('.', origCursor)

    return result
endfunc
"}}}

function! omnicpp#scopes#GetLocalUsingNamespaces() "{{{2
    return s:GetLocalUsingNamespaces()
endfunc
"}}}
" 获取当前光标所在的 {} 块中的 using namespace
function! s:GetLocalUsingNamespaces() "{{{2
    let result = []

    " 在局部作用域中搜索. 反向搜索
    let origCursor = getpos('.')
    let origPos = origCursor[1:2]

    let stopPos = omnicpp#scopes#GetCurBlockStartPos()

    let stopLine = stopPos[0]
    let curPos = origPos[:]
    while 1
		" 搜索局部与某作用域(非作用于整个源文件)的 'using namespace' 声明
        let curPos = searchpos('\Cusing\s\+namespace', 'bW', stopLine)

        " 搜索失败
        if curPos == [0, 0]
            break
        elseif omnicpp#scopes#GetCurBlockStartPos() != stopPos
            " 停留在其他的 {} 块中, 跳过
            continue
        endif

        let szLine = getline('.')

        " eg. using namespace MyNamesp|ace;
        if curPos[0] == origPos[0]
            " We get the line until cursor position
            " 只找到当前光标所在行的一个语句
            " 直接跳过, 因为可能语句不完整
            " TODO: 检测有效的声明
            continue
            "let szLine = szLine[: origPos[1]-1]
        endif

        let szLine = omni#cpp#utils#GetCodeFromLine(szLine)
        " We get the namespace list from the line
        let result = s:GetNamespaceListFromLine(szLine) + result
    endwhile

    " Setting the cursor to the original position
    call setpos('.', origCursor)

    return result
endfunc
"}}}

" Get global using namespace list from the current buffer
" stopLine 搜索在此行后停止. 搜索时同样搜索 stopLine
function! s:GetGlobalUsingNamespaces(stopLine) "{{{2
    " FIXME: using namespace MyNamespa|ce;
    let result = []
    let origCursor = getpos('.')

    call setpos('.', [0, 1, 1, 0])
    let curPos = [1,1]
    while curPos != [0,0]
		"FIXME: 如果此语句不是在行首呢？过于复杂，暂时为简单的实现
        let curPos = searchpos('\C^using\s\+namespace', 'W', a:stopLine)
        if curPos != [0,0]
            let szLine = getline('.')
            let startPos = curPos[1]
            let endPos = match(szLine, ';', startPos-1)
            if endPos!=-1
                " We get the namespace list from the line
                let result += s:GetNamespaceListFromLine(szLine)
            endif
        endif
    endwhile

    call setpos('.', origCursor)
    return result
endfunc
"}}}

" Get the using namespace list from a line
" 从字符串中提取使用的名空间列表
function! s:GetNamespaceListFromLine(szLine) "{{{2
    let result = []
    let tokens = omnicpp#tokenizer#Tokenize(a:szLine)
    let szNamespace = ''
    let state = 0
    for token in tokens
        if state==0
            let szNamespace = ''
            if token.value == '/*'
                let state = 1
            elseif token.value == '//'
                " It's a comment
                let state = -1
                break
            elseif token.value == 'using'
                let state = 2
            endif
        elseif state==1
            if token.value == '*/'
                let state=0
            endif
        elseif state==2
            if token.value == 'namespace'
                let state = 3
            else
                " Error, 'using' must be followed by 'namespace'
                let state = -1
                break
            endif
        elseif state==3
            if token.value == '::'
                let szNamespace .= token.value
                let state = 4
            elseif token.kind == 'cppWord'
                let szNamespace .= token.value
                let state = 5
                " Maybe end of tokens
            endif
        elseif state==4
            if token.kind == 'cppWord'
                let szNamespace .= token.value
                let state = 5
                " Maybe end of tokens
            else
                " Error, we can't have 'using namespace Something::'
                let state = -1
                break
            endif
        elseif state==5
            if token.value == '::'
                let szNamespace .= token.value
                let state = 4
            else
                call extend(result, [szNamespace])
                let state = 0
            endif
        endif
    endfor

    if state == 5
        call extend(result, [szNamespace])
    endif

    return result
endfunc
"}}}

" Get the stop position when searching for local variables
" 搜索当前块的开始位置
" TODO: 处理 extern "C" {
function! s:GetStopPositionForLocalSearch() "{{{2
    " Stop position when searching a local variable
    let originalPos = getpos('.')
    let origPos = originalPos[1:2]
    let stopPosition = origPos
    let curPos = origPos
    while curPos != [0, 0]
        let stopPosition = curPos
        let curPos = searchpairpos('{', '', '}', 'bW', 
                    \g:omnicpp#utils#expIgnoreComments)
    endwhile
    call setpos('.', originalPos)

    return stopPosition
endfunc
"}}}

" 返回当前 {} 块的开始位置
" 若传入参数且非 0, 会移动光标至开始位置
" 若不在 {} 块中, 返回 [0, 0]
function! omnicpp#scopes#GetCurBlockStartPos(...) "{{{2
    if a:0 > 0 && a:1
        let szFlag = 'bW'
    else
        let szFlag = 'bWn'
    endif
    let startPos = searchpairpos('{', '', '}', szFlag, 
                \g:omnicpp#utils#expIgnoreComments)
    return startPos
endfunc
"}}}
" 返回当前 {} 块的结束位置
" 若传入参数且非 0, 会移动光标至开始位置
" 若不在 {} 块中, 返回 [0, 0]
function! omnicpp#scopes#GetCurBlockEndPos(...) "{{{2
    if a:0 > 0 && a:1
        let szFlag = 'W'
    else
        let szFlag = 'Wn'
    endif
    let startPos = searchpairpos('{', '', '}', szFlag, 
                \g:omnicpp#utils#expIgnoreComments)
    return startPos
endfunc
"}}}

" vim:fdm=marker:fen:expandtab:smarttab:fdl=1:
autoload/omnicpp/settings.vim	[[[1
78
" Description:  Omnicpp completion init settings
" Maintainer:   fanhe <fanhed@163.com>
" Last Change:  2011 May 15
" License:      GPLv2

function! omnicpp#settings#Init()
    " Sets the namespace search method
    "   0 = disabled
    "   1 = search namespaces in the current file
    "   2 = search namespaces in the current file and included files
    if !exists('g:VLOmniCpp_NamespaceSearch') 
        let g:VLOmniCpp_NamespaceSearch = 1
    endif

    " Set the class scope completion mode
    "   0 = auto
    "   1 = show all members (static, public, protected and private)
    if !exists('g:VLOmniCpp_DisplayMode') 
        let g:VLOmniCpp_DisplayMode = 0
    endif

    " Set if the parent is displayed in the abbr column of the popup
    "   0 = no
    "   1 = yes
    if !exists('g:VLOmniCpp_ShowParentInAbbr') 
        let g:VLOmniCpp_ShowParentInAbbr = 0
    endif

    " Set if the access (+,#,-) is displayed
    "   0 = no
    "   1 = yes
    if !exists('g:VLOmniCpp_ShowAccess') 
        let g:VLOmniCpp_ShowAccess = 1
    endif

    " Set the list of default namespaces
    " eg: ['std']
    if !exists('g:VLOmniCpp_DefaultNamespaces') 
        let g:VLOmniCpp_DefaultNamespaces = []
    endif

    " Set MayComplete to '.'
    "   0 = disabled
    "   1 = enabled
    "   default = 1
    if !exists('g:VLOmniCpp_MayCompleteDot') 
        let g:VLOmniCpp_MayCompleteDot = 1
    endif

    " Set MayComplete to '->'
    "   0 = disabled
    "   1 = enabled
    "   default = 1
    if !exists('g:VLOmniCpp_MayCompleteArrow') 
        let g:VLOmniCpp_MayCompleteArrow = 1
    endif

    " Set MayComplete to '::'
    "   0 = disabled
    "   1 = enabled
    "   default = 2
    if !exists('g:VLOmniCpp_MayCompleteColon') 
        let g:VLOmniCpp_MayCompleteColon = 1
    endif

    " When completeopt does not contain longest option, this setting 
    " controls the behaviour of the popup menu selection 
    " when starting the completion
    "   0 = don't select first item
    "   1 = select first item (inserting it to the text)
    "   2 = select first item (without inserting it to the text)
    "   default = 2
    if !exists('g:VLOmniCpp_ItemSelectionMode') 
        let g:VLOmniCpp_ItemSelectionMode = 2
    endif
endfunc

" vim:fdm=marker:fen:expandtab:smarttab:fdl=1:
autoload/omnicpp/tokenizer.vim	[[[1
124
" Description:  Omnicpp completion tokenizer
" Maintainer:   fanhe <fanhed@163.com>
" Last Change:  2011 May 11
" License:      GPLv2

" TODO: Generic behaviour for Tokenize()


" C++ 的关键词列表
" From the C++ BNF
let s:cppKeyword = ['asm', 'auto', 'bool', 'break', 'case', 'catch', 'char', 
            \'class', 'const', 'const_cast', 'continue', 'default', 'delete', 
            \'do', 'double', 'dynamic_cast', 'else', 'enum', 'explicit', 
            \'export', 'extern', 'false', 'float', 'for', 'friend', 'goto', 
            \'if', 'inline', 'int', 'long', 'mutable', 'namespace', 'new', 
            \'operator', 'private', 'protected', 'public', 'register', 
            \'reinterpret_cast', 'return', 'short', 'signed', 'sizeof', 
            \'static', 'static_cast', 'struct', 'switch', 'template', 'this', 
            \'throw', 'true', 'try', 'typedef', 'typeid', 'typename', 'union', 
            \'unsigned', 'using', 'virtual', 'void', 'volatile', 'wchar_t', 
            \'while', 'and', 'and_eq', 'bitand', 'bitor', 'compl', 'not', 
            \'not_eq', 'or', 'or_eq', 'xor', 'xor_eq']

" 用于匹配 C++ 关键词的正则表达式
let s:reCppKeyword = '\C\<'.join(s:cppKeyword, '\>\|\<').'\>'

" The order of items in this list is very important because we use this list to 
" build a regular expression (see below) for tokenization
" C++ 操作符和标点符号列表, 顺序非常重要, 因为这用于生成用于标记化的正则表达式
let s:cppOperatorPunctuator = ['->*', '->', '--', '-=', '-', '!=', '!', '##', 
            \'#', '%:%:', '%=', '%>', '%:', '%', '&&', '&=', '&', '(', ')', 
            \'*=', '*', ',', '...', '.*', '.', '/=', '/', '::', ':>', ':', 
            \';', '?', '[', ']', '^=', '^', '{', '||', '|=', '|', '}', '~', 
            \'++', '+=', '+', '<<=', '<%', '<:', '<<', '<=', '<', '==', '=', 
            \'>>=', '>>', '>=', '>']

" We build the regexp for the tokenizer
" 用于标记化的正则表达式
let s:reCComment = '\/\*\|\*\/'
let s:reCppComment = '\/\/'
let s:reComment = s:reCComment.'\|'.s:reCppComment
let s:reCppOperatorOrPunctuator = escape(
            \join(s:cppOperatorPunctuator, '\|'), '*./^~[]')


" Tokenize a c++ code
" a token is dictionary where keys are:
"   -   kind    = cppKeyword|cppWord|cppOperatorPunctuator|cComment|
"                 cppComment|cppDigit|unknown
"   -   value   = <text>
"   Note: a cppWord is any word that is not a cpp keyword
"   Note: 不能解析的, 被忽略. 中文会解析错误.
"   TODO: 忽略注释
function! omnicpp#tokenizer#Tokenize(szCode) "{{{2
    let result = []

    " The regexp to find a token, a token is a keyword, word or
    " c++ operator or punctuator. To work properly we have to put 
    " spaces and tabs to our regexp.
    let reTokenSearch = '\(\w\+\)\|\s\+\|'.s:reComment.'\|'
                \.s:reCppOperatorOrPunctuator
    " eg: 'using namespace std;'
    "      ^    ^
    "  start=0 end=5
    let startPos = 0
    let endPos = matchend(a:szCode, reTokenSearch)
    let len = endPos - startPos
    while endPos != -1
        " eg: 'using namespace std;'
        "      ^    ^
        "  start=0 end=5
        "  token = 'using'
        " We also remove space and tabs
        let token = substitute(strpart(a:szCode, startPos, len), '\s', '', 'g')

        " eg: 'using namespace std;'
        "           ^         ^
        "       start=5     end=15
        let startPos = endPos
        let endPos = matchend(a:szCode, reTokenSearch, startPos)
        let len = endPos - startPos

        " It the token is empty we continue
        if token == ''
            continue
        endif

        " Building the token
        let resultToken = {'kind' : 'unknown', 'value' : token}

        " Classify the token
        if token =~ '^\d\+'
            " It's a digit
            let resultToken.kind = 'cppDigit'
        elseif token =~'^\w\+$'
            " It's a word
            let resultToken.kind = 'cppWord'

            " But maybe it's a c++ keyword
            if match(token, s:reCppKeyword) >= 0
                let resultToken.kind = 'cppKeyword'
            endif
        else
            if match(token, s:reComment) >= 0
                if index(['/*','*/'],token) >= 0
                    let resultToken.kind = 'cComment'
                else
                    let resultToken.kind = 'cppComment'
                endif
            else
                " It's an operator
                let resultToken.kind = 'cppOperatorPunctuator'
            endif
        endif

        " We have our token, let's add it to the result list
        call extend(result, [resultToken])
    endwhile

    return result
endfunc


" vim:fdm=marker:fen:expandtab:smarttab:fdl=1:
autoload/omnicpp/TagEntry.py	[[[1
630
#!/usr/bin/env python
# -*- encoding:utf-8 -*-


CKinds = {
    'c': "class",     
    'd': "macro",     
    'e': "enumerator",
    'f': "function",  
    'g': "enum",      
    'l': "local",     
    'm': "member",    
    'n': "namespace", 
    'p': "prototype", 
    's': "struct",    
    't': "typedef",   
    'u': "union",     
    'v': "variable",  
    'x': "externvar", 
}

RevCKinds = {
    "class":        'c',
    "macro":        'd',
    "enumerator":   'e',
    "function":     'f',
    "enum":         'g',
    "local":        'l',
    "member":       'm',
    "namespace":    'n',
    "prototype":    'p',
    "struct":       's',
    "typedef":      't',
    "union":        'u',
    "variable":     'v',
    "externvar":    'x',
}

def ToFullKind(kind):
    result = kind
    if CKinds.has_key(kind):
        result = CKinds[kind]
    return result
def ToFullKinds(kinds):
    return [ ToFullKind(kind) for kind in kinds ]


def ToAbbrKind(kind):
    result = kind
    if RevCKinds.has_key(kind):
        result = RevCKinds[kind]
    return result
def ToAbbrKinds(kinds):
    return [ ToAbbrKind(kind) for kind in kinds ]

def GetMacroArgList():
    #TODO
    pass

class TagEntry():
    def __init__(self):
        self.name = ''              # Tag name (short name, excluding any scope
                                        # names)
        self.file = ''              # File this tag is found

        self.pattern = ''           # A pattern that can be used to locate the 
                                        # tag in the file

        self.kind = '<unknown>'     # Member, function, class, typedef etc.
        self.lineNumber = -1        # Line number
        self.scope = ''             # Scope

        self.qualifiers = ''        # Qualifiers. 
                                    # eg1. int a, b, c; -> int
                                    # eg2. void * Func() -> void *
                                    # eg3. template<typename T>

        self.path = ''              # Tag full path
        self.parent = ''            # Direct parent
        self.parentType = ''        # Parent type eg. namespace, class, enum,
                                        # struct, union

        self.extFields = {}         # Additional extension fields

        self.hti = None             # Handle to tree item, not persistent item
        self.id = -1
        self.differOnByLineNumber = False

    def __eq__(self, rhs):
        res2 = self.scope == rhs.scope \
                and self.file == rhs.file \
                and self.kind == rhs.kind \
                and self.parent == rhs.parent \
                and self.pattern == rhs.pattern \
                and self.name == rhs.name \
                and self.path == rhs.path \
                and self.GetInheritsAsString() == rhs.GetInheritsAsString() \
                and self.GetAccess() == rhs.GetAccess() \
                and self.GetSignature() == rhs.GetSignature() \
                and self.GetTyperef() == rhs.GetTyperef()
        res = res2 and self.lineNumber == rhs.lineNumber

        if res2 and not res:
            # the entries are differs only in the line numbers
            self.differOnByLineNumber = True

        return res

    def Create(self, fileName, name, lineNumber, pattern, kind, extFields):
        self.SetName(name)
        self.SetLine(lineNumber)
        if kind:
            self.SetKind(kind)
        else:
            self.SetKind('<unknown>')
        self.SetPattern(pattern)
        self.SetFile(fileName)
        self.SetId(-1)
        self.extFields = extFields

        # ======================================================================
        # 从 pattern 提取 qualifiers, 以后可能会改变
        tmpPattern = self.GetPattern()[2:-2].strip()
        if tmpPattern.startswith('template'):
            qualifiers = ''
            if self.IsClass():
                qualifiers = tmpPattern.split('class')[0].strip()
            elif self.IsFunction():
                qualifiers = tmpPattern.partition('(')[0].rstrip()\
                        .rpartition(' ')[0].strip()
            self.SetQualifiers(qualifiers)
        # 从 returns 提取 qualifiers, 以后可能会改变
        if self.GetReturnValue():
            self.SetQualifiers(self.GetReturnValue().strip())

        # ======================================================================

        # Check if we can get full name (including path)
        # 添加 parentType 属性, 以保证不丢失信息
        path = self.GetExtField('class')
        if path:
            self.SetParentType('class')
            self.UpdatePath(path)
        else:
            path = self.GetExtField('struct')
            if path:
                self.SetParentType('struct')
                self.UpdatePath(path)
            else:
                path = self.GetExtField('namespace')
                if path:
                    self.SetParentType('namespace')
                    self.UpdatePath(path)
                else:
                    path = self.GetExtField('interface')
                    if path:
                        self.SetParentType('interface')
                        self.UpdatePath(path)
                    else:
                        path = self.GetExtField('union')
                        if path:
                            self.SetParentType('union')
                            self.UpdatePath(path)
                        #FIXME: 真的需要剔除 __anon 字样?
                        #tmpName = path.rpartition(':')[2]
                        #if path:
                            #self.SetParentType('union')
                            #if not tmpName.startswith('__anon'):
                                #self.UpdatePath(path)
                            #else:
                                ## anonymous union, 
                                ## remove the anonymous part from its name
                                #path = path.rpartition(':')[2]
                                #path = path.rpartition(':')[2]
                                #self.UpdatePath(path)
        if path:
            self.SetScope(path)
        else:
            self.SetScope('<global>')

        # If there is no path, path is set to name
        if not self.GetPath():
            self.SetPath(self.GetName())

        # Get the parent name
        tok = self.GetPath().split('::')
        if len(tok) < 2:
            parent = '<global>'
        else:
            parent = tok[-2]
        self.SetParent(parent)

    def FromLine(self, line):
        strLine = line
        lineNumber = -1
        extFields = {}

        # get the token name
        partStrList = strLine.partition('\t')
        name = partStrList[0]
        strLine = partStrList[2]

        # get the file name
        partStrList = strLine.partition('\t')
        fileName = partStrList[0]
        strLine = partStrList[2]

        # here we can get two options:
        # pattern followed by ;"
        # or
        # line number followed by ;"
        partStrList = strLine.partition(';"')
        if not partStrList[1]:
            # invalid pattern found
            return

        if strLine.startswith('/^'):
            # regular expression pattern found
            pattern = partStrList[0]
            strLine = partStrList[2]
        else:
            # line number pattern found, this is usually the case when
            # dealing with macros in C++
            pattern = partStrList[0].strip()
            strLine = partStrList[2]
            lineNumber = int(pattern)

        # next is the kind of the token
        if strLine.startswith('\t'):
            strLine = strLine.lstrip('\t')

        partStrList = strLine.partition('\t')
        kind = partStrList[0]
        strLine = partStrList[2]

        if strLine:
            for i in strLine.split('\t'):
                key = i.partition(':')[0].strip()
                val = i.partition(':')[2].strip()

                if key == 'line' and val:
                    lineNumber = int(val)
                else:
                    if key == 'union' or key == 'struct':
                        # remove the anonymous part of the struct / union
                        # FIXME: 不应该删除无名结构/联合的标志
                        #if not val.startswith('__anon'):
                            #scopeArr = val.split(':')
                            #tmp = ''
                            #for j in scopeArr:
                                #if not j:
                                    # 分离的标志应该改为 ::
                                    #continue
                                #if not j.startswith('__anon'):
                                    #tmp += j + '::'
                            #val = tmp.rstrip('::')

                        # 不删除任何信息
                        pass

                    extFields[key] = val

        # 真的需要?
        kind = kind.strip()
        name = name.strip()
        fileName = fileName.strip()
        pattern = pattern.strip()

        if kind == 'enumerator':
            # enums are specials, they are not really a scope so they should 
            # appear when I type: enumName::
            # they should be member of their parent 
            # (which can be <global>, or class)
            # but we want to know the "enum" type they belong to, 
            # so save that in typeref,
            # then patch the enum field to lift the enumerator into the 
            # enclosing scope.
            # watch out for anonymous enums -- leave their typeref field blank.
            if extFields.has_key('enum'):
                typeref = extFields['enum']
                extFields['enum'] = \
                        extFields['enum'].rpartition(':')[0].rpartition(':')[0]
                if not typeref.rpartition(':')[2].startswith('__anon'):
                    # watch out for anonymous enums
                    # just leave their typeref field blank.
                    extFields['typeref'] = typeref

        self.Create(fileName, name, lineNumber, pattern, kind, extFields)

    def IsOk(self):
        return self.GetKind() != "<unknown>"

    def IsContainer(self):
        kind = self.GetKind()
        return kind == 'class' \
                or kind == 'struct' \
                or kind == 'union' \
                or kind == 'namespace' \
                or kind == 'project'

    def IsConstructor(self):
        if self.GetKind() != 'function' and self.GetKind() != 'prototype':
            return False
        else:
            return self.GetName() == self.GetScope()

    def IsDestructor(self):
        if self.GetKind() != 'function' and self.GetKind() != 'prototype':
            return False
        else:
            return self.GetName().startswith('~')

    def IsMethod(self):
        '''Return true of the this tag is a function or prototype'''
        return self.IsPrototype() or self.IsFunction()

    def IsFunction(self):
        return self.GetKind() == 'function'

    def IsPrototype(self):
        return self.GetKind() == 'prototype'

    def IsMacro(self):
        return self.GetKind() == 'macro'

    def IsClass(self):
        return self.GetKind() == 'class'

    def IsStruct(self):
        return self.GetKind() == 'struct'

    def IsScopeGlobal(self):
        return not self.GetScope() or self.GetScope() == '<global>'

    def IsTypedef(self):
        return self.GetKind() == 'typedef'


	#------------------------------------------
	# Operations
	#------------------------------------------
    def GetDifferOnByLineNumber(self):
        return self.differOnByLineNumber

    def GetId(self):
        return self.id
    def SetId(self, id):
        self.id = id

    def GetName(self):
        return self.name
    def SetName(self, name):
        self.name = name

    def GetPath(self):
        return self.path
    def SetPath(self, path):
        self.path = path

    def GetFile(self):
        return self.file
    def SetFile(self, file):
        self.file = file

    def GetLine(self):
        return self.lineNumber
    def SetLine(self, line):
        self.lineNumber = line

    def GetPattern(self):
        # since ctags's pattern is regex, forward slashes are escaped. 
        # ('/' becomes '\/')
        pattern = self.pattern.replace('\\\\', '\\').replace('\\/', '/')

        # TODO: 只有 pattern 可能包含非 ascii 字符?
        if not isinstance(pattern, unicode):
            try:
                import chardet
                encoding = chardet.detect(pattern)['encoding']
                if encoding != 'utf-8' and encoding != 'ascii':
                    try:
                        pattern = pattern.decode(encoding).encode('utf-8')
                    except LookupError:
                        # FIXME: 编码不支持?!
                        pattern = 'LookupError: '+ encoding
                    except UnicodeDecodeError:
                        pattern = 'UnicodeDecodeError: ' + encoding
            except ImportError:
                pass
        return pattern

    def SetPattern(self, pattern):
        self.pattern = pattern

    def GetKind(self):
        return self.kind
    def GetAbbrKind(self):
        return RevCKinds[self.kind]
    def SetKind(self, kind):
        # 保存的类型必须为全名
        if len(kind) == 1:
            kind = CKinds[kind]
        self.kind = kind.strip()

    def GetParent(self):
        return self.parent
    def SetParent(self, parent):
        self.parent = parent

    def GetParentType(self):
        return self.parentType
    def SetParentType(self, parentType):
        self.parentType = parentType

    def GetQualifiers(self):
        return self.qualifiers
    def SetQualifiers(self, qualifiers):
        self.qualifiers = qualifiers

    def GetAccess(self):
        return self.GetExtField("access")
    def SetAccess(self, access):
        self.extFields["access"] = access

    def GetSignature(self):
        return self.GetExtField("signature")
    def SetSignature(self, sig):
        self.extFields["signature"] = sig

    def SetInherits(self, inherits):
        self.extFields["inherits"] = inherits
    def GetInherits(self):
        return self.GetInheritsAsString()

    def GetTyperef(self):
        return self.GetExtField("typeref")
    def SetTyperef(self, typeref):
        self.extFields["typeref"] = typeref

    def GetInheritsAsString(self):
        return self.GetExtField('inherits')

    def GetInheritsAsArrayNoTemplates(self):
        '''返回清除了模版信息的继承字段的列表'''
        inherits = self.GetInheritsAsString()
        parent = ''
        parentsArr = []

        # 清楚所有尖括号内的字符串
        depth = 0
        for ch in inherits:
            if ch == '<':
                if depth == 0 and parent:
                    parentsArr.append(parent.strip())
                    parent = ''
                depth += 1
            elif ch == '>':
                depth -= 1
            elif ch == ',':
                if depth == 0 and parent:
                    parentsArr.append(parent.strip())
                    parent = ''
            else:
                if depth == 0:
                    parent += ch

        if parent:
            parentsArr.append(parent.strip())

        return parentsArr

    def GetInheritsAsArrayWithTemplates(self):
        inherits = self.GetInheritsAsString()
        parent = ''
        parentsArr = []

        depth = 0
        for ch in inherits:
            if ch == '<':
                depth += 1
                parent += ch
            elif ch == '>':
                depth -= 1
                parent += ch
            elif ch == ',':
                if depth == 0 and parent:
                    parentsArr.append(parent.strip())
                    parent = ''
                elif depth != 0:
                    parent += ch
            else:
                parent += ch

        if parent:
            parentsArr.append(parent.strip())

        return parentsArr

    def GetReturnValue(self):
        returnValue = self.GetExtField('returns').strip().replace('virtual', '')
        return returnValue

    def SetReturnValue(self, retVal):
        self.extFields["returns"] = retVal

    def GetScope(self):
        return self.scope
    def SetScope(self, scope):
        self.scope = scope

    def GetScopeName(self):
        '''Return scope name of the tag.

        If path is empty in db or contains just the project name, 
        it will return the literal <global>.
        For project tags, an empty string is returned.
        '''
        return self.GetScope()

    def Key(self):
        '''Generate a Key for this tag based on its attributes

        Return tag key'''
        # 键值为 [原型/宏:]path:signature
        key = ''
        if self.GetKind == 'prototype' or self.GetKind() == 'macro':
            key += self.GetKind() + ': '

        key += self.GetPath() + self.GetSignature()
        return key

    def GetDisplayName(self):
        '''Generate a display name for this tag to be used by the symbol tree

        Return tag display name'''
        return self.GetName() + self.GetSignature()

    def GetFullDisplayName(self):
        '''Generate a full display name for this tag that includes:
        full scope + name + signature

        Return tag full display name
        '''
        name = ''
        if self.GetParent() == '<global>':
            name += self.GetDisplayName()
        else:
            name += self.GetParent() + '::' + self.GetName() \
                    + self.GetSignature()
        return name

    def NameFromTyperef(self, templateInitList):
        '''Return the actual name as described in the 'typeref' field

        Return real name or wxEmptyString'''
        #TODO
        pass

    def TypeFromTyperef(self):
        '''Return the actual type as described in the 'typeref' field

        return real name or wxEmptyString'''
        typeref = self.GetTyperef()
        if typeref:
            name = typeref.partition(':')[0]
            return name
        else:
            return ''

	# ------------------------------------------
	#  Extenstion fields
	# ------------------------------------------
    def GetExtField(self, extField):
        if self.extFields.has_key(extField):
            return self.extFields[extField]
        else:
            return ''

	# ------------------------------------------
	#  Misc
	# ------------------------------------------
    def Print(self):
        '''顺序基本与数据库的一致'''
        print '======================================'
        print 'Name:\t\t' + self.GetName()
        print 'File:\t\t' + self.GetFile()
        print 'Line:\t\t' + str(self.GetLine())
        print 'Kind:\t\t' + self.GetKind()
        print 'Pattern:\t' + self.GetPattern()
        print 'Parent:\t\t' + self.GetParent()
        print 'ParentType:\t\t' + self.GetParentType()
        print 'Path:\t\t' + self.GetPath()
        print 'Scope:\t\t' + self.GetScope()
        print 'Qualifiers:\t\t' + self.GetQualifiers()
        print ' ---- Ext fields: ---- '
        for k, v in self.extFields.iteritems():
            print k + ':\t\t' + v
        print '======================================'

    def ReplaceSimpleMacro(self):
        #TODO
        pass

    def UpdatePath(self, path):
        '''Update the path with full path (e.g. namespace::class)'''
        if path:
            name = path
            name += '::'
            name += self.GetName()
            self.SetPath(name)

    def TypedefFromPattern(self, tagPattern, typedefName, name, templateInit):
        '''从 tags 中的 pattern 字段提取 typedef'''
        pattern = tagPattern.lstrip('/^')
        #TODO: 需要 lex...
        pass


if __name__ == '__main__':
    FILE = 'cltagssort'
    FILE = 'ctagsWithClOpts.tags'
    FILE = 'TagsTest/tags'
    with open(FILE) as f:
        for line in f.readlines():
            if line.startswith('!'):
                continue
            entry = TagEntry()
            entry.FromLine(line)
            entry.Print()

autoload/omnicpp/includes.vim	[[[1
233
" Description:  Omni completion script for resolve incldue files
" Maintainer:   fanhe <fanhed@163.com>
" Last Change:  2011 May 13
" License:      GPLv2

" Note: 若需添加自定义的头文件搜索路径, 设置 &path 变量(:h 'path')

" TODO: 不适用 vim 的选项变量. 缺点, 不能 set += 
let g:Omnicpp_IncludeSearchPaths = &path


"{{{2 内部用变量
" 文件包含列表缓存, 格式为 {<filePath> : <incList>}
" <incList> = {'include': <filePath>, 'pos': <pos>}
let s:CACHE_INCLUDES = {}
" 文件修改时间缓存, 格式为 {<filePath> : <mtime>}
let s:CACHE_FILEMTIME = {}

let s:rePreprocIncludePart = '\C^\s*#\s*include\s*'
let s:reIncludeFilePart = '\(<\|"\)\(\f\|\s\)\+\(>\|"\)'
let s:rePreprocIncludeFile = s:rePreprocIncludePart . s:reIncludeFilePart

"}}}
" 返回当前缓冲区包含的文件列表(嵌套解析)
" 具体路径根据 &path 变量的值而定
function! omnicpp#includes#GetIncludeFiles() "{{{2
    " 用字典键值保存结果以防止重复
    let includeGuard = {}
    let includeResult = [] " 另外的用于保存结果的列表, 有点, 保持顺序
    let szCurFilePath = expand('%:p')
    call s:DoGetIncludeFiles(szCurFilePath, includeGuard, includeResult)

    " 第一项为当前文件, 删除之
    call remove(includeResult, 0)
    " 也需要处理缓存
    if has_key(s:CACHE_INCLUDES, szCurFilePath)
        call remove(s:CACHE_INCLUDES, szCurFilePath)
        call remove(s:CACHE_FILEMTIME, szCurFilePath)
    endif
    let b:CACHE_INCLUDES = s:CACHE_INCLUDES
    let b:CACHE_FILEMTIME = s:CACHE_FILEMTIME

    return includeResult
endfunc


function! s:DoGetIncludeFiles(szFilePath, includeGuard, includeResult) "{{{2
    let includeGuard = a:includeGuard
    let includeResult = a:includeResult
    let szFilePath = omnicpp#includes#ResolveFilePath(a:szFilePath)
    if has_key(includeGuard, szFilePath)
        return
    else
        " includeGuard 的键值也就是解析完成的包含文件
        let includeGuard[szFilePath] = 1
        " includeResult 为结果, 不会重复, 保持包含的顺序
        call add(includeResult, szFilePath)
    endif

    let incList = omnicpp#includes#GetIncludeList(a:szFilePath)
    for inc in incList
        call s:DoGetIncludeFiles(inc.include, includeGuard, includeResult)
    endfor
endfunc


" Resolve the path of the file according to the &path variable
" 根据 &path 变量扩展文件路径
" 返回绝对路径
function! omnicpp#includes#ResolveFilePath(szFile) "{{{2
    let result = ''
    let listPath = split(globpath(&path, a:szFile), "\n")
    if len(listPath)
        let result = fnamemodify(listPath[0], ':p')
    endif
    " fnamemodify() 貌似已经执行了路径简化
    "return fnamemodify(simplify(result), ':p')
    return result
endfunc


" Get the include list of a file
" Note: 没有找到的头文件不会返回
" Param1: 空为表示获取当前缓冲区, 否则为获取指定文件
function! omnicpp#includes#GetIncludeList(...) "{{{2
    if a:0 > 0
        return s:GetIncludeListFromFile(a:1)
    else
        return s:GetIncludeListFromCurrentBuffer()
    endif
endfunc


" Get the include list from the current buffer
function! s:GetIncludeListFromCurrentBuffer() "{{{2
    let listIncludes = []
    let originalPos = getpos('.')

    call setpos('.', [0, 1, 1, 0])
    let curPos = [1,1]
    let alreadyInclude = {} " 键值保存包含的文件. 防止重复.
    while curPos != [0,0]
        let curPos = searchpos('\C\('.s:rePreprocIncludeFile.'\)', 'W')
        if curPos != [0,0]
            let szLine = getline('.')
            let startPos = curPos[1]
            let endPos = matchend(szLine, s:reIncludeFilePart, startPos-1)
            if endPos!=-1
                let szInclusion = szLine[startPos-1:endPos-1]
                let szIncludeFile = substitute(
                            \szInclusion, 
                            \'\('.s:rePreprocIncludePart.'\)\|[<>""]', '', 'g')
                let szResolvedInclude = 
                            \omnicpp#includes#ResolveFilePath(szIncludeFile)

                " Protection over self inclusion
                " 防止循环包含, 并且令列表中的项不重复
                if szResolvedInclude != '' 
                            \&& szResolvedInclude 
                            \   != omnicpp#includes#ResolveFilePath(getreg('%'))
                    let includePos = curPos
                    if !has_key(alreadyInclude, szResolvedInclude)
                        call extend(listIncludes, 
                                    \[{'pos' : includePos, 
                                    \   'include' : szResolvedInclude}])
                        let alreadyInclude[szResolvedInclude] = 1
                    endif
                endif
            endif
        endif
    endwhile

    call setpos('.', originalPos)
    return listIncludes
endfunc


" Get the include list from a file
function! s:GetIncludeListFromFile(szFilePath) "{{{2
    let listIncludes = []
    if a:szFilePath == ''
        return listIncludes
    endif

    let szAbsFilePath = fnamemodify(a:szFilePath, ':p')

    if has_key(s:CACHE_INCLUDES, szAbsFilePath)
        " 比较时间戳, 如果磁盘的文件修改时间不比缓存的新, 直接从缓存获取
        if getftime(szAbsFilePath) 
                    \<= s:CACHE_FILEMTIME[szAbsFilePath]
            return copy(s:CACHE_INCLUDES[szAbsFilePath])
        endif
    endif

    " 保持两个字典的键值同步
    let s:CACHE_INCLUDES[szAbsFilePath] = []
    let s:CACHE_FILEMTIME[szAbsFilePath] = getftime(szAbsFilePath)

    let szFixedPath = escape(szAbsFilePath, g:omnicpp#utils#szEscapedCharacters)
    execute 'silent! lvimgrep /\C\('.s:rePreprocIncludeFile.'\)/gj '.szFixedPath

    let listQuickFix = getloclist(0)
    let alreadyInclude = {}
    for qf in listQuickFix
        let szLine = qf.text
        let startPos = qf.col
        let endPos = matchend(szLine, s:reIncludeFilePart, startPos-1)
        if endPos!=-1
            let szInclusion = szLine[startPos-1:endPos-1]
            let szIncludeFile = substitute(
                        \szInclusion, 
                        \'\('.s:rePreprocIncludePart.'\)\|[<>""]', '', 'g')
            let szResolvedInclude 
                        \= omnicpp#includes#ResolveFilePath(szIncludeFile)
            
            " Protection over self inclusion
            " 防止循环包含, 并且令列表中的项不重复
            if szResolvedInclude != '' && szResolvedInclude != szAbsFilePath
                let includePos = [qf.lnum, qf.col]
                if !has_key(alreadyInclude, szResolvedInclude)
                    call extend(listIncludes, 
                                \[{'pos' : includePos, 
                                \   'include' : szResolvedInclude}])
                    let alreadyInclude[szResolvedInclude] = 1
                endif
            endif
        endif
    endfor

    " 缓存结果
    let s:CACHE_INCLUDES[szAbsFilePath] = listIncludes

    return copy(listIncludes)
endfunc


function! omnicpp#includes#ClearCache() "清空缓存, 用于强制更新全部 {{{2
    " 一般用在修改 &path 或者 getcwd() 后
    call filter(s:CACHE_INCLUDES, 0)
    call filter(s:CACHE_FILEMTIME, 0)
endfunc


" 返回当前缓冲区包含的文件(包括嵌套的, 忽略不存在的)
" For debug purpose
function! omnicpp#includes#Display() "{{{2
    let szPathBuffer = omnicpp#includes#ResolveFilePath(getreg('%'))
    call s:DisplayIncludeTree(szPathBuffer, 0)
endfunc


" For debug purpose
function! s:DisplayIncludeTree(szFilePath, indent, ...) "{{{2
    let includeGuard = {}
    if a:0 >0
        let includeGuard = a:1
    endif
    let szFilePath = omnicpp#includes#ResolveFilePath(a:szFilePath)
    if has_key(includeGuard, szFilePath)
        return
    else
        let includeGuard[szFilePath] = 1
    endif

    let szIndent = repeat('    ', a:indent)
    echom szIndent . a:szFilePath
    let incList = omnicpp#includes#GetIncludeList(a:szFilePath)
    for inc in incList
        call s:DisplayIncludeTree(inc.include, a:indent+1, includeGuard)
    endfor
endfunc

" vim:fdm=marker:fen:expandtab:smarttab:fdl=1:
autoload/omnicpp/FileEntry.py	[[[1
29
#!/usr/bin/env python
# -*- encoding:utf-8 -*-


class FileEntry:
    def __init__(self):
        self.id = -1
        self.file = ''
        self.lastRetaggedTimestamp = 0 # 可用当前时间初始化

    def SetFile(self, file):
        self.file = file

    def SetLastRetaggedTimestamp(self, lastRetaggedTimestamp):
        self.lastRetaggedTimestamp = lastRetaggedTimestamp

    def GetFile(self):
        return self.file

    def GetLastRetaggedTimestamp(self):
        return self.lastRetaggedTimestamp

    def SetId(self, id):
        self.id = id

    def GetId(self):
        return self.id


autoload/omnicpp/test.vim	[[[1
390
"echo omnicpp#resolvers#GetOmniScopeStack('A::B C::D::')
"echo omnicpp#resolvers#GetOmniScopeStack('A::B::C().')
"echo omnicpp#resolvers#GetOmniScopeStack('A::B().C->')
"echo omnicpp#resolvers#GetOmniScopeStack('A->B().')
"echo omnicpp#resolvers#GetOmniScopeStack('A->B.')
"echo omnicpp#resolvers#GetOmniScopeStack('Z Y = ((A*)B)->C.')
"echo omnicpp#resolvers#GetOmniScopeStack('(A*)B()->C.')
"

function! s:GetCurStatementStartPos() "{{{2
    let origCursor = getpos('.')
    let origPos = origCursor[1:2]

    let result = [0, 0]

    while searchpos('[;{}]\|\%^', 'bW') != [0, 0]
        "if omnicpp#utils#IsCursorInCommentOrString()
            "continue
        "else
            break
        "endif
    endwhile

    if getpos('.')[1:2] == [1, 1] " 到达文件头的话, 接受光标所在位置的匹配
        let result = searchpos('[^ \t]', 'Wc')
    else
        let result = searchpos('[^ \t]', 'W')
    endif

    call setpos('.', origCursor)

    return result
endfunc
function! g:TestSyntaxSpeed()
    let nTimes = 100
    call g:Timer.Start()
    for i in range(nTimes)
        call omnicpp#utils#GetCurStatementStartPos()
    endfor
    call g:Timer.EndEchoMes()
    call g:Timer.Start()
    for i in range(nTimes)
        call s:GetCurStatementStartPos()
    endfor
    call g:Timer.EndEchoMes()
endfunction


" 简化 omni 补全请求代码, 效率不是太高
" eg1. A::B("\"")->C(" a(\")z ", ')'). -> A::B()->C().
" eg2. ((A*)(B))->C. -> ((A*)B)->C.
" egx. A::B(C.D(), ((E*)F->G)).
function! g:SimplifyOmniCode(szOmniCode) "{{{2
    let s = a:szOmniCode
    
    " 1. 清理函数函数
    let s = s:StripFuncArgs(s)
    " 2. 清理多余的括号, 不会检查匹配的情况. eg. ((A*)((B))) -> ((A*)B)
    " TODO: 需要更准确
    let s = substitute(s, '\W\zs(\+\(\w\+\))\+\ze', '\1', 'g')

    return s
endfunc
"}}}
" 剔除所有函数参数, 仅保留一个括号
" TODO: 处理引号中的 ')' 干扰
function! s:StripFuncArgs(szOmniCode) "{{{2
    let s = a:szOmniCode

    " 1. 清理引号内的 \" , 因为正则貌似是不能直接匹配 ("\"")
    let s = substitute(s, '\\"', '', 'g')
    " 2. 清理双引号的内容, 否则双引号内的括号会干扰括号替换
    let s = substitute(s, '".\{-}"', '', 'g')
    " 3. 清理 '(' 和 ')', 避免干扰
    let s = substitute(s, "'('\\|')'", '', 'g')

    let szResult = ''
    let nStart = 0
    while nStart < len(s) && nStart != -1
        let nEnd = matchend(s, '\w\s*(', nStart)
        if nEnd != -1
            let szResult .= s[nStart : nEnd - 1]

            let nStart = nEnd
            let nCount = 1
            " 开始寻找匹配的 ) 的位置
            for i in range(nStart, len(s) - 1)

                let c = s[i]

                if c == '('
                    let nCount += 1
                elseif c == ')'
                    let nCount -= 1
                endif

                if nCount == 0
                    let nStart = i + 1
                    let szResult .= ')'
                    break
                endif
            endfor

        else
            let szResult .= s[nStart :]
            break
        endif
    endwhile

    return szResult
endfunc
"}}}

"echo g:SimplifyOmniCode('A::B("\"")->C(" a(\")z ", '')'').')
"echo g:SimplifyOmniCode('((A*)(B))->C.')

"echo g:StripFuncArgs('A::B(C.D(), ((E*)F->G)).')

"echo g:StripFuncArgs('A::B(C.D(), ((E*)F->G)).H("1, 2)", '')'', Z()->Y()->X)->')


function! g:TestGetCurBlockStartPos()
    call g:Timer.Start()
    let nTimes = 100
    for i in range(nTimes)
        call omnicpp#scopes#GetCurBlockStartPos()
    endfor
    call g:Timer.EndEchoMes()
endfunc

function! g:TestGetCurBlockEndPos()
    call g:Timer.Start()
    let nTimes = 100
    for i in range(nTimes)
        call omnicpp#scopes#GetCurBlockEndPos()
    endfor
    call g:Timer.EndEchoMes()
endfunc
function! g:TestGetCurBlockEndPos2()
    call g:Timer.Start()
    let nTimes = 100
    for i in range(nTimes)
        let startPos = searchpairpos('{', '', '}', 'Wn')
    endfor
    call g:Timer.EndEchoMes()
endfunc
function! g:TestGetCurBlockEndPos3()
    call g:Timer.Start()
    let nTimes = 100
    for i in range(nTimes)
        let startPos = g:Searchpairpos()
    endfor
    call g:Timer.EndEchoMes()
endfunc
function! g:TestGetCurBlockEndPos4()
    call g:Timer.Start()
    let nTimes = 100
    for i in range(nTimes)
        call searchpair('{', '', '}', 'Wn', 
                    \'synIDattr(synID(line("."), col("."), 0), "name") =~? "string"')
    endfor
    call g:Timer.EndEchoMes()
endfunc
function! g:TestGetCurBlockEndPos5()
    call g:Timer.Start()
    let nTimes = 100
    for i in range(nTimes)
        let origCursor = getpos('.')
        call searchpair('{', '', '}', 'Wb')
        normal! %
        let b:startPos = getpos('.')[1:2]
        call setpos('.', origCursor)
    endfor
    call g:Timer.EndEchoMes()
endfunc
function! g:Searchpairpos()
    let origCursor = getpos('.')
    let nCount = 1
    while nCount > 0
        call search('{\|}')
        if getline('.')[col('.')-1] == '{'
            let nCount += 1
        else
            let nCount -= 1
        endif
    endwhile
    let result = getpos('.')[1:2]
    call setpos('.', origCursor)
    return result
endfunction



function! s:NewTypeInfo() "{{{2
    return {'name': '', 'tsdm': {}, 'tsl': [], 'tdl': []}
endfunc
"}}}
function! s:StripString(s) "{{{2
    let s = substitute(a:s, '^\s\+', '', 'g')
    let s = substitute(s, '\s\+$', '', 'g')
    return s
endfunc
"}}}
" Param: s  为模版声明或定义字符串
" Return:   模版初始化列表
function! s:GetTemplateSpecList(s) "{{{2
    let s = a:s
    let til = []

    let idx = 0
    let nCount = 0
    let s1 = 0
    while idx < len(s)
        let c = s[idx]
        if c == '<'
            let nCount += 1
            if nCount == 1
                let s1 = idx + 1
            endif
        elseif c == '>'
            let nCount -= 1
            if nCount == 0
                call add(til, s:StripString(s[s1 : idx-1]))
                break
            endif
        elseif c == ','
            if nCount == 1
                call add(til, s:StripString(s[s1 : idx-1]))
                let s1 = idx + 1
            endif
        endif

        let idx += 1
    endwhile

    return til
endfunc
"}}}
function! g:GetVariableTypeInfo(szDecl, szVar) "{{{2
    let typeInfo = s:NewTypeInfo()
    " FIXME: python 的字典转为 vim 字典时, \t 全部丢失
    let szDecl = substitute(a:szDecl, '\\t', ' ', 'g')
    let s = szDecl[: match(szDecl, a:szVar)-1]
    " 删除赋值语句
    let s = substitute(s, '=.*$', '', 'g')
    " eg. int *a, *b, *| -> int |
    let s = substitute(s, '\%(>\|\w\)\zs\s\+\%(\*\+\|&\)\?\s*\w\+\s*,.*$', '', 
                \'g')
    if s =~# '>\s*$'
        " 有模版
        let typeInfo.name = matchstr(s, '\w\+\ze\s*<.\+$')

        " 提取模版初始化列表
        let tds = matchstr(s, '\w\+\s*\zs<.\+$')
        let typeInfo.tdl = s:GetTemplateSpecList(tds)
    else
        let typeInfo.name = matchstr(s, '\w\+\ze\s*[*&]\?\s*$')
    endif

    return typeInfo
endfunc
"}}}

"echo g:GetVariableTypeInfo(' B< A, C < char, short > > **c1, c2, c3;', 'c3')
"echo g:GetVariableTypeInfo(' B * c1, & c2 = c1, c3;', 'c3')
"echo g:GetVariableTypeInfo(' B & c1, & c2 = c1, c3;', 'c3')
"echo g:GetVariableTypeInfo(' B c1, & c2 = c1, c3;', 'c3')
"echo g:GetVariableTypeInfo(' A < B, C , D < E, F > , G<H,I,K<L,M,N> > > x', 'x')

"echo s:GetTemplateSpecList(' A < B, C , D < E , F > , G<H,I,K<L,M,N> > >')

" 剔除配对的字符里面(包括配对的字符)的内容
function! s:StripPair(szString, szStart, szEnd, ...) "{{{2
    let szString = a:szString
    let szStart = a:szStart
    let szEnd = a:szEnd
    let nDeep = 1
    if a:0 > 0
        let nDeep = a:1
    endif

    if nDeep <= 0
        return szString
    endif

    let result = ''

    let idx = 0
    let nLen = len(szString)
    let nCount = 0
    let nSubStart = 0
    while idx < nLen
        let c = szString[idx]
        if c == szStart
            let nCount += 1
            if nCount == nDeep && nSubStart != -1
                let result .= szString[nSubStart : idx - 1]
                let nSubStart = -1
            endif
        elseif c == szEnd
            let nCount -= 1
            if nCount == nDeep - 1
                let nSubStart = idx + 1
            endif
        endif
        let idx += 1
    endwhile

    if nSubStart != -1
        let result .= szString[nSubStart : -1]
    endif

    return result
endfunc
"}}}

"echo s:StripPair('B1<A1>,B2<A1,T<a,b,c<d,e<f>>>>,B3', '<', '>')
"echo s:StripPair('B1<A1>,B2<A1,T<a,b,c<d,e<f>>>>,B3', '<', '>', 2)
"echo s:StripPair('A(B(C(D))), E(), F(G)', '(', ')', 0)
"echo s:StripPair('A(B(C(D))), E(), F(G)', '(', ')', 1)
"echo s:StripPair('A(B(C(D))), E(), F(G)', '(', ')', 2)
"echo s:StripPair('A(B(C(D))), E(), F(G)', '(', ')', 3)
"echo s:StripPair('A(B(C(D))), E(), F(G)', '(', ')', 4)


function! s:GetTemplateDeclList(sQualifiers) "{{{2
    let text = a:sQualifiers
    if text == ''
        return []
    endif

    let text = substitute(text, '^.\+<', '', 'g')
    let text = substitute(text, '>.*$', '', 'g')
    let text = substitute(text, 'typename', '', 'g')
    let text = substitute(text, 'class', '', 'g')

    return map(split(text, ','), 's:StripString(v:val)')
endfunc
"}}}

function! s:GetInheritsInfoList(sInherits) "{{{2
    let sInherits = a:sInherits
    if sInherits == ''
        return []
    endif

    let lResult = []

    let idx = 0
    let nNestDepth = 0
    let nLen = len(sInherits)
    let nSubStart = 0
    while idx < nLen
        let c = sInherits[idx]
        if c == '<'
            let nNestDepth += 1
        elseif c == '>'
            let nNestDepth -= 1
        elseif c == ','
            if nNestDepth == 0
                let sText = sInherits[nSubStart : idx - 1]
                let dInfo = {}
                let dInfo.name = matchstr(sText, '^\s*\zs\w\+\ze\s*<\?')
                let dInfo.tsl = s:GetTemplateSpecList(sText)
                call add(lResult, dInfo)
                let nSubStart = idx + 1
            endif
        endif

        let idx += 1
    endwhile

    let sText = sInherits[nSubStart : idx - 1]
    let dInfo = {}
    let dInfo.name = matchstr(sText, '^\s*\zs\w\+\ze\s*<\?')
    let dInfo.tsl = s:GetTemplateSpecList(sText)
    call add(lResult, dInfo)
    let nSubStart = idx + 1

    return lResult
endfunc
"}}}

"echo s:GetInheritsInfoList('B1<A1>,B2<A1,T>,B3,B4<A2<a,b>, A3<c,d<e>>>')
"echo s:GetInheritsInfoList('B1,B2')
"echo s:GetInheritsInfoList('B1<A1>,B2<A1,T>')

"echo s:GetTemplateDeclList(' template < T1 , T2 , T3> const T3 & ')

" vim:fdm=marker:fen:expandtab:smarttab:fdl=1:
autoload/omnicpp/ITagsStorage.py	[[[1
62
#!/usr/bin/env python
# -*- encoding:utf-8 -*-



class ITagsStorage:
    '''ITagsStorage defined the tags storage API used by vimlite'''
    OrderNone = 0
    OrderAsc = 1
    OrderDesc = 2

    TagOk = 10
    TagExist = 11
    TagError = 12

    def __init__(self):
        self.fileName = ''
        self.singleSearchLimit = 1000
        self.maxWorkspaceTagToColour = 1000
        self.useCache = False

    def SetUseCache(self, useCache):
        self.useCache = useCache

    def GetUseCache(self):
        return self.useCache

    def GetDatabaseFileName(self):
        return self.fileName

    def SetSingleSearchLimit(self, singleSearchLimit):
        if singleSearchLimit < 0:
            singleSearchLimit = 1000
        self.singleSearchLimit = singleSearchLimit

    def GetSingleSearchLimit(self):
        return self.singleSearchLimit

    def SetMaxWorkspaceTagToColour(self, maxWorkspaceTagToColour):
        self.maxWorkspaceTagToColour = maxWorkspaceTagToColour

    def GetMaxWorkspaceTagToColour(self):
        return self.maxWorkspaceTagToColour





class StorageCacheEnabler:
    '''helper class to turn on/off storage cache flag'''
    def __init__(self, storage):
        self.storage = storage
        if self.storage:
            self.storage.SetUseCache(True)

    def __del__(self):
        if self.storage:
            self.storage.SetUseCache(False)
            self.storage = None



